"""liquidctl buses and drivers.

The typical use case of generic scripts and interfaces – including the
liquidctl CLI – is to instantiate drivers for all known devices found on the
system.

    from liquidctl.driver import *
    for dev in find_liquidctl_devices():
        print(dev.description)

Is also possible to find devices compatible with a specific driver.

    from liquidctl.driver.kraken_two import KrakenTwoDriver
    for dev in KrakenTwoDriver.find_supported_devices():
        print(dev.description)

Copyright Jonas Malaco and contributors
SPDX-License-Identifier: GPL-3.0-or-later
"""

import sys

from liquidctl.driver.base import BaseBus, find_all_subclasses

# automatically enabled drivers
from liquidctl.driver import asetek
from liquidctl.driver import asetek_pro
from liquidctl.driver import aura_led
from liquidctl.driver import aquacomputer
from liquidctl.driver import commander_core
from liquidctl.driver import commander_pro
from liquidctl.driver import corsair_hid_psu
from liquidctl.driver import hydro_platinum
from liquidctl.driver import kraken2
from liquidctl.driver import kraken3
from liquidctl.driver import nzxt_epsu
from liquidctl.driver import rgb_fusion2
from liquidctl.driver import smart_device
if sys.platform == 'linux':
    from liquidctl.driver import ddr4
    from liquidctl.driver import nvidia


def find_liquidctl_devices(pick=None, **kwargs):
    """Find devices and instantiate corresponding liquidctl drivers.

    Probes all buses and drivers that have been loaded at the time of the call
    and yields driver instances.

    Filter conditions can be passed through to the buses and drivers via
    `**kwargs`.  A driver instance will be yielded for each compatible device
    that matches the supplied filter conditions.

    If `pick` is passed, only the driver instance for the `(pick + 1)`-th
    matched device will be yielded.
    """
    buses = sorted(find_all_subclasses(BaseBus),
                   key=lambda x: (x.__module__, x.__name__))
    num = 0
    for bus_cls in buses:
        for dev in bus_cls().find_devices(**kwargs):
            if pick is not None:
                if num == pick:
                    yield dev
                    return
                num += 1
            else:
                yield dev


__all__ = [
    'find_liquidctl_devices',
]

# allow old driver imports to continue to work by manually placing these into
# the module cache, so import liquidctl.driver.foo does not need to
# check the filesystem for foo
sys.modules['liquidctl.driver.kraken_two'] = kraken2
sys.modules['liquidctl.driver.nzxt_smart_device'] = smart_device
sys.modules['liquidctl.driver.seasonic'] = nzxt_epsu



import pytest
from _testutils import MockHidapiDevice, Report, MockRuntimeStorage

from liquidctl.driver.commander_pro import _quoted, _prepare_profile, _fan_mode_desc, CommanderPro
from liquidctl.driver.hwmon import HwmonDevice
from liquidctl.error import NotSupportedByDevice


# hardcoded responce data expected for some of the calls:
# commander pro: firmware request (0.9.214)
# commander pro: bootloader req (2.3)
# commander pro: get temp config ( 3 sensors)
# commander pro: get fan configs (3 DC fans, 1 PWM fan) # note I have not tested it with pwm fans
# commander pro:


@pytest.fixture
def commanderProDeviceUnconnected():
    device = MockHidapiDevice(vendor_id=0x1b1c, product_id=0x0c10, address='addr')
    return CommanderPro(device, 'Corsair Commander Pro', 6, 4, 2)


@pytest.fixture
def lightingNodeProDeviceUnconnected():
    device = MockHidapiDevice(vendor_id=0x1b1c, product_id=0x0c0b, address='addr')
    return CommanderPro(device, 'Corsair Lighting Node Pro', 0, 0, 2)


@pytest.fixture
def commanderProDevice():
    device = MockHidapiDevice(vendor_id=0x1b1c, product_id=0x0c10, address='addr')
    pro = CommanderPro(device, 'Corsair Commander Pro', 6, 4, 2)


    runtime_storage = MockRuntimeStorage(key_prefixes=['testing'])
    pro.connect(runtime_storage=runtime_storage)
    return pro


@pytest.fixture
def lightingNodeProDevice():
    device = MockHidapiDevice(vendor_id=0x1b1c, product_id=0x0c0b, address='addr')
    node = CommanderPro(device, 'Corsair Lighting Node Pro', 0, 0, 2)
    runtime_storage = MockRuntimeStorage(key_prefixes=['testing'])
    node.connect(runtime_storage=runtime_storage)
    return node


@pytest.fixture
def lightingNodeCoreDevice():
    device = MockHidapiDevice(vendor_id=0x1b1c, product_id=0x0c1a, address='addr')
    node = CommanderPro(device, 'Corsair Lighting Node Core', 0, 0, 1)
    runtime_storage = MockRuntimeStorage(key_prefixes=['testing'])
    node.connect(runtime_storage=runtime_storage)
    return node


# prepare profile
def test_prepare_profile_valid_max_rpm():
    assert _prepare_profile([[10, 400], [20, 5000]], 60) == [(10, 400), (20, 5000), (60, 5000), (60, 5000), (60, 5000), (60, 5000)]


def test_prepare_profile_add_max_rpm():
    assert _prepare_profile([[10, 400]], 60) == [(10, 400), (60, 5000), (60, 5000), (60, 5000), (60, 5000), (60, 5000)]
    assert _prepare_profile([[10, 400], [20, 500], [30, 600], [40, 700], [50, 800]], 60) == [(10, 400), (20, 500), (30, 600), (40, 700), (50, 800), (60, 5000)]


def test_prepare_profile_missing_max_rpm():
    with pytest.raises(ValueError):
        _prepare_profile([[10, 400], [20, 500], [30, 600], [40, 700], [50, 800], [55, 900]], 60)


def test_prepare_profile_full_set():
    assert _prepare_profile([[10, 400], [20, 500], [30, 600], [40, 700], [45, 2000], [50, 5000]], 60) == [(10, 400), (20, 500), (30, 600), (40, 700), (45, 2000), (50, 5000)]


def test_prepare_profile_too_many_points():
    with pytest.raises(ValueError):
        _prepare_profile([[10, 400], [20, 500], [30, 600], [40, 700], [50, 800], [55, 900]], 60)


def test_prepare_profile_no_points():
    assert _prepare_profile([], 60) == [(60, 5000), (60, 5000), (60, 5000), (60, 5000), (60, 5000), (60, 5000)]


def test_prepare_profile_empty_list():
    assert _prepare_profile([], 60) == [(60, 5000), (60, 5000), (60, 5000), (60, 5000), (60, 5000), (60, 5000)]


def test_prepare_profile_above_max_temp():
    assert _prepare_profile([[10, 400], [70, 2000]], 60) == [(10, 400), (60, 5000), (60, 5000), (60, 5000), (60, 5000), (60, 5000)]


def test_prepare_profile_temp_low():
    assert _prepare_profile([[-10, 400], [70, 2000]], 60) == [(-10, 400), (60, 5000), (60, 5000), (60, 5000), (60, 5000), (60, 5000)]


def test_prepare_profile_max_temp():
    assert _prepare_profile([], 100) == [(100, 5000), (100, 5000), (100, 5000), (100, 5000), (100, 5000), (100, 5000)]


# quoted
def test_quoted_empty():
    assert _quoted() == ''


def test_quoted_single():
    assert _quoted('one arg') == "'one arg'"


def test_quoted_valid():
    assert _quoted('one', 'two') == "'one', 'two'"


def test_quoted_not_string():
    assert _quoted('test', 500) == "'test', 500"


# fan modes
def test_get_fan_mode_description_auto():
    assert _fan_mode_desc(0x00) == None


def test_get_fan_mode_description_unknown():
    assert _fan_mode_desc(0x03) == None
    assert _fan_mode_desc(0x04) == None
    assert _fan_mode_desc(0x10) == None
    assert _fan_mode_desc(0xff) == None


def test_get_fan_mode_description_dc():
    assert _fan_mode_desc(0x01) == 'DC'


def test_get_fan_mode_description_pwm():
    assert _fan_mode_desc(0x02) == 'PWM'


# class methods
def test_commander_constructor(commanderProDeviceUnconnected):

    assert commanderProDeviceUnconnected._data is None
    assert commanderProDeviceUnconnected._fan_names == ['fan1', 'fan2', 'fan3', 'fan4', 'fan5', 'fan6']
    assert commanderProDeviceUnconnected._led_names == ['led1', 'led2']
    assert commanderProDeviceUnconnected._temp_probs == 4
    assert commanderProDeviceUnconnected._fan_count == 6


def test_lighting_constructor(lightingNodeProDeviceUnconnected):
    assert lightingNodeProDeviceUnconnected._data is None
    assert lightingNodeProDeviceUnconnected._fan_names == []
    assert lightingNodeProDeviceUnconnected._led_names == ['led1', 'led2']
    assert lightingNodeProDeviceUnconnected._temp_probs == 0
    assert lightingNodeProDeviceUnconnected._fan_count == 0


def test_connect_commander(commanderProDeviceUnconnected):
    commanderProDeviceUnconnected.connect()
    assert commanderProDeviceUnconnected._data is not None


def test_connect_lighting(lightingNodeProDeviceUnconnected):
    lightingNodeProDeviceUnconnected.connect()
    assert lightingNodeProDeviceUnconnected._data is not None


@pytest.mark.parametrize('exp,fan_modes', [([(3, 0x01)], {'4': 'dc'}), ([(3, 0x00)], {'4': 'off'}), ([(4, 0x02)], {'5': 'pwm'}), ([(0, 0x01),(2, 0x02),(1, 0x01)], {'1': 'dc', '3':'pwm', '2': 'dc'})])
def test_initialize_commander_pro_fan_modes(commanderProDevice, exp, fan_modes, tmp_path):

    responses = [
        '000009d4000000000000000000000000',  # firmware
        '00000500000000000000000000000000',  # bootloader
        '00010100010000000000000000000000',  # temp probes
        '00010102000000000000000000000000',  # fan set (throw away)
        '00010102000000000000000000000000',  # fan set (throw away)
        '00010102000000000000000000000000',  # fan set (throw away)
        '00010102000000000000000000000000'   # fan probes
    ]
    for d in responses:
        commanderProDevice.device.preload_read(Report(0, bytes.fromhex(d)))

    commanderProDevice.initialize(direct_access=True, fan_modes=fan_modes)

    sent = commanderProDevice.device.sent
    assert len(sent) == 4+len(exp)

    for i in range(len(exp)):
        assert sent[3+i].data[0] == 0x28
        assert sent[3+i].data[2] == exp[i][0]
        assert sent[3+i].data[3] == exp[i][1]


@pytest.mark.parametrize('has_hwmon,direct_access', [(False, False), (True, True), (True, False)])
def test_initialize_commander_pro(commanderProDevice, has_hwmon, direct_access, tmp_path):
    if has_hwmon and not direct_access:
        commanderProDevice._hwmon = HwmonDevice('mock_module', tmp_path)
        (tmp_path / 'temp1_input').write_text('10000\n')
        (tmp_path / 'temp2_input').write_text('20000\n')
        (tmp_path / 'temp4_input').write_text('40000\n')
        (tmp_path / 'fan1_input').write_text('1100\n')
        (tmp_path / 'fan1_label').write_text('fan1 3pin\n')
        (tmp_path / 'fan2_input').write_text('1200\n')
        (tmp_path / 'fan2_label').write_text('fan1 3pin\n')
        (tmp_path / 'fan3_input').write_text('1300\n')
        (tmp_path / 'fan3_label').write_text('fan1 4pin\n')
    else:
        if has_hwmon:
            commanderProDevice._hwmon = HwmonDevice('invalid_module', None)
        responses = [
            '000009d4000000000000000000000000',  # firmware
            '00000500000000000000000000000000',  # bootloader
            '00010100010000000000000000000000',  # temp probes
            '00010102000000000000000000000000'   # fan probes
        ]
        for d in responses:
            commanderProDevice.device.preload_read(Report(0, bytes.fromhex(d)))

    res = commanderProDevice.initialize(direct_access=direct_access)

    if has_hwmon and not direct_access:
        assert len(res) == 10
        i = 0
    else:
        assert len(res) == 12
        assert res[0] == ('Firmware version', '0.9.212', '')
        assert res[1] == ('Bootloader version', '0.5', '')
        i = 2

    assert res[i + 0] == ('Temperature probe 1', True, '')
    assert res[i + 1] == ('Temperature probe 2', True, '')
    assert res[i + 2] == ('Temperature probe 3', False, '')
    assert res[i + 3] == ('Temperature probe 4', True, '')

    assert res[i + 4] == ('Fan 1 control mode', 'DC', '')
    assert res[i + 5] == ('Fan 2 control mode', 'DC', '')
    assert res[i + 6] == ('Fan 3 control mode', 'PWM', '')
    assert res[i + 7] == ('Fan 4 control mode', None, '')
    assert res[i + 8] == ('Fan 5 control mode', None, '')
    assert res[i + 9] == ('Fan 6 control mode', None, '')

    data = commanderProDevice._data.load('fan_modes', None)
    assert data is not None
    assert len(data) == 6
    assert data[0] == 0x01
    assert data[1] == 0x01
    assert data[2] == 0x02
    assert data[3] == 0x00
    assert data[4] == 0x00
    assert data[5] == 0x00

    data = commanderProDevice._data.load('temp_sensors_connected', None)
    assert data is not None
    assert len(data) == 4
    assert data[0] == 0x01
    assert data[1] == 0x01
    assert data[2] == 0x00
    assert data[3] == 0x01


def test_initialize_lighting_node(lightingNodeProDevice):
    responses = [
        '000009d4000000000000000000000000',  # firmware
        '00000500000000000000000000000000'   # bootloader
    ]
    for d in responses:
        lightingNodeProDevice.device.preload_read(Report(0, bytes.fromhex(d)))

    res = lightingNodeProDevice.initialize()

    assert len(res) == 2
    assert res[0][1] == '0.9.212'
    assert res[1][1] == '0.5'

    data = lightingNodeProDevice._data.load('fan_modes', None)
    assert data is None

    data = lightingNodeProDevice._data.load('temp_sensors_connected', None)
    assert data is None


@pytest.mark.parametrize('has_hwmon,direct_access', [(False, False), (True, True), (True, False)])
def test_get_status_commander_pro(commanderProDevice, has_hwmon, direct_access, tmp_path):
    if has_hwmon and not direct_access:
        commanderProDevice._hwmon = HwmonDevice('mock_module', tmp_path)
        (tmp_path / 'temp1_input').write_text('26910\n')
        (tmp_path / 'temp2_input').write_text('29220\n')
        (tmp_path / 'temp4_input').write_text('25740\n')
        (tmp_path / 'fan1_input').write_text('940\n')
        (tmp_path / 'fan1_label').write_text('fan1 3pin\n')
        (tmp_path / 'fan2_input').write_text('939\n')
        (tmp_path / 'fan2_label').write_text('fan1 3pin\n')
        (tmp_path / 'fan3_input').write_text('987\n')
        (tmp_path / 'fan3_label').write_text('fan1 4pin\n')
        (tmp_path / 'in0_input').write_text('12066\n')
        (tmp_path / 'in1_input').write_text('4965\n')
        (tmp_path / 'in2_input').write_text('3359\n')
    else:
        if has_hwmon:
            commanderProDevice._hwmon = HwmonDevice('invalid_module', None)
        responses = [
            '000a8300000000000000000000000000',  # temp sensor 1
            '000b6a00000000000000000000000000',  # temp sensor 2
            '000a0e00000000000000000000000000',  # temp sensor 4
            '0003ac00000000000000000000000000',  # fan speed 1
            '0003ab00000000000000000000000000',  # fan speed 2
            '0003db00000000000000000000000000',  # fan speed 3
            '002f2200000000000000000000000000',  # get 12v
            '00136500000000000000000000000000',  # get 5v
            '000d1f00000000000000000000000000',  # get 3.3v
        ]
        for d in responses:
            commanderProDevice.device.preload_read(Report(0, bytes.fromhex(d)))

    commanderProDevice._data.store('fan_modes', [0x01, 0x01, 0x02, 0x00, 0x00, 0x00])
    commanderProDevice._data.store('temp_sensors_connected', [0x01, 0x01, 0x00, 0x01])

    res = commanderProDevice.get_status(direct_access=direct_access)
    print(res)

    assert len(res) == 9

    # temp probes
    assert res[0] == ('Temperature 1', pytest.approx(26.91, abs=1e-3), '°C')
    assert res[1] == ('Temperature 2', pytest.approx(29.22, abs=1e-3), '°C')
    assert res[2] == ('Temperature 4', pytest.approx(25.74, abs=1e-3), '°C')

    # fans rpm
    assert res[3] == ('Fan 1 speed', 940, 'rpm')
    assert res[4] == ('Fan 2 speed', 939, 'rpm')
    assert res[5] == ('Fan 3 speed', 987, 'rpm')

    # voltages
    assert res[6] == ('+12V rail', pytest.approx(12.066, abs=1e-3), 'V')
    assert res[7] == ('+5V rail', pytest.approx(4.965, abs=1e-3), 'V')
    assert res[8] == ('+3.3V rail', pytest.approx(3.359, abs=1e-3), 'V')

    if not has_hwmon or direct_access:
        # check the commands sent
        sent = commanderProDevice.device.sent
        assert len(sent) == 9

        assert sent[0].data[0] == 0x11
        assert sent[1].data[0] == 0x11
        assert sent[2].data[0] == 0x11

        assert sent[3].data[0] == 0x21
        assert sent[4].data[0] == 0x21
        assert sent[5].data[0] == 0x21

        assert sent[6].data[0] == 0x12
        assert sent[7].data[0] == 0x12
        assert sent[8].data[0] == 0x12


def test_get_status_lighting_pro(lightingNodeProDevice):

    res = lightingNodeProDevice.get_status()
    assert len(res) == 0


def test_get_temp_valid_sensor_commander(commanderProDevice):

    response = '000a8300000000000000000000000000'
    commanderProDevice.device.preload_read(Report(0, bytes.fromhex(response)))

    commanderProDevice._data.store('temp_sensors_connected', [0x01, 0x01, 0x01, 0x01])

    res = commanderProDevice._get_temp(1)

    assert res == 26.91

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 1
    assert sent[0].data[0] == 0x11
    assert sent[0].data[1] == 1


def test_get_temp_invalid_sensor_low_commander(commanderProDevice):
    response = '000a8300000000000000000000000000'
    commanderProDevice.device.preload_read(Report(0, bytes.fromhex(response)))

    commanderProDevice._data.store('temp_sensors_connected', [0x01, 0x01, 0x01, 0x01])

    with pytest.raises(ValueError):
        commanderProDevice._get_temp(-1)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 0


def test_get_temp_invalid_sensor_high_commander(commanderProDevice):
    response = '000a8300000000000000000000000000'
    commanderProDevice.device.preload_read(Report(0, bytes.fromhex(response)))

    commanderProDevice._data.store('temp_sensors_connected', [0x01, 0x01, 0x01, 0x01])

    with pytest.raises(ValueError):
        commanderProDevice._get_temp(4)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 0


def test_get_temp_lighting(lightingNodeProDevice):
    response = '000a8300000000000000000000000000'
    lightingNodeProDevice.device.preload_read(Report(0, bytes.fromhex(response)))

    lightingNodeProDevice._data.store('temp_sensors_connected', [0x00, 0x00, 0x00, 0x00])

    with pytest.raises(ValueError):
        lightingNodeProDevice._get_temp(2)

    # check the commands sent
    sent = lightingNodeProDevice.device.sent
    assert len(sent) == 0


def test_get_fan_rpm_valid_commander(commanderProDevice):

    response = '0003ac00000000000000000000000000'
    commanderProDevice.device.preload_read(Report(0, bytes.fromhex(response)))
    commanderProDevice._data.store('fan_modes', [0x01, 0x01, 0x02, 0x00, 0x00, 0x00])

    res = commanderProDevice._get_fan_rpm(1)
    assert res == 940

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 1
    assert sent[0].data[0] == 0x21
    assert sent[0].data[1] == 1


def test_get_fan_rpm_invalid_low_commander(commanderProDevice):

    response = '0003ac00000000000000000000000000'
    commanderProDevice.device.preload_read(Report(0, bytes.fromhex(response)))

    commanderProDevice._data.store('fan_modes', [0x01, 0x01, 0x02, 0x00, 0x00, 0x00])

    with pytest.raises(ValueError):
        commanderProDevice._get_fan_rpm(-1)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 0


def test_get_fan_rpm_invalid_high_commander(commanderProDevice):
    response = '0003ac00000000000000000000000000'
    commanderProDevice.device.preload_read(Report(0, bytes.fromhex(response)))

    commanderProDevice._data.store('fan_modes', [0x01, 0x01, 0x02, 0x00, 0x00, 0x00])

    with pytest.raises(ValueError):
        commanderProDevice._get_fan_rpm(7)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 0


def test_get_fan_rpm_lighting(lightingNodeProDevice):
    response = '0003ac00000000000000000000000000'
    lightingNodeProDevice.device.preload_read(Report(0, bytes.fromhex(response)))

    with pytest.raises(ValueError):
        lightingNodeProDevice._get_fan_rpm(7)

    # check the commands sent
    sent = lightingNodeProDevice.device.sent
    assert len(sent) == 0


def test_get_hw_fan_channels_all(commanderProDevice):

    res = commanderProDevice._get_hw_fan_channels('sync')
    assert res == [0, 1, 2, 3, 4, 5]


def test_get_hw_fan_channels_lowercase(commanderProDevice):
    res = commanderProDevice._get_hw_fan_channels('fan2')
    assert res == [1]


@pytest.mark.parametrize('channel', [
    'fan23', 'fan7', 'fan0', 'fan', 'led', 'led1', 'bob'
    ])
def test_get_hw_fan_channels_invalid(commanderProDevice, channel):
    with pytest.raises(ValueError):
        commanderProDevice._get_hw_fan_channels(channel)


@pytest.mark.parametrize('channel,expected', [
    ('led1', [0]), ('led2', [1]), ('sync', [0, 1])
    ])
def test_get_hw_led_channel_valid(commanderProDevice, channel, expected):
    res = commanderProDevice._get_hw_led_channels(channel)
    assert res == expected


@pytest.mark.parametrize('channel,expected', [('led', [0]), ('any', [0])])
def test_get_hw_led_channel_valid_node_core(lightingNodeCoreDevice, channel,
                                            expected):
    res = lightingNodeCoreDevice._get_hw_led_channels(channel)
    assert res == expected


@pytest.mark.parametrize('channel', [
    'led0', 'led3', 'led', 'fan', 'fan1', 'bob'
    ])
def test_get_hw_led_channels_invalid(commanderProDevice, channel):
    with pytest.raises(ValueError):
        commanderProDevice._get_hw_led_channels(channel)


def test_set_fixed_speed_low(commanderProDevice):

    response = '00000000000000000000000000000000'
    commanderProDevice.device.preload_read(Report(0, bytes.fromhex(response)))
    commanderProDevice._data.store('fan_modes', [0x01, 0x01, 0x01, 0x01, 0x01, 0x01])

    commanderProDevice.set_fixed_speed('fan4', -10)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 1

    assert sent[0].data[0] == 0x23
    assert sent[0].data[1] == 0x03
    assert sent[0].data[2] == 0x00


def test_set_fixed_speed_high(commanderProDevice):
    response = '00000000000000000000000000000000'
    commanderProDevice.device.preload_read(Report(0, bytes.fromhex(response)))
    commanderProDevice._data.store('fan_modes', [0x01, 0x01, 0x01, 0x01, 0x01, 0x01])

    commanderProDevice.set_fixed_speed('fan3', 110)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 1

    assert sent[0].data[0] == 0x23
    assert sent[0].data[1] == 0x02
    assert sent[0].data[2] == 0x64


def test_set_fixed_speed_valid(commanderProDevice):

    response = '00000000000000000000000000000000'
    commanderProDevice.device.preload_read(Report(0, bytes.fromhex(response)))
    commanderProDevice._data.store('fan_modes', [0x01, 0x01, 0x01, 0x01, 0x01, 0x01])

    commanderProDevice.set_fixed_speed('fan2', 50)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 1

    assert sent[0].data[0] == 0x23
    assert sent[0].data[1] == 0x01
    assert sent[0].data[2] == 0x32


def test_set_fixed_speed_valid_unconfigured(commanderProDevice):

    response = '00000000000000000000000000000000'
    commanderProDevice.device.preload_read(Report(0, bytes.fromhex(response)))
    commanderProDevice._data.store('fan_modes', [0x00, 0x00, 0x00, 0x00, 0x00, 0x00])

    commanderProDevice.set_fixed_speed('fan2', 50)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 0


def test_set_fixed_speed_valid_multi_fan(commanderProDevice):
    responses = [
        '00000000000000000000000000000000',
        '00000000000000000000000000000000',
        '00000000000000000000000000000000',
        '00000000000000000000000000000000',
        '00000000000000000000000000000000'
    ]

    for d in responses:
        commanderProDevice.device.preload_read(Report(0, bytes.fromhex(d)))

    commanderProDevice._data.store('fan_modes', [0x01, 0x00, 0x01, 0x01, 0x00, 0x00])

    commanderProDevice.set_fixed_speed('sync', 50)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 3

    assert sent[0].data[0] == 0x23
    assert sent[0].data[1] == 0x00
    assert sent[0].data[2] == 0x32

    assert sent[1].data[0] == 0x23
    assert sent[1].data[1] == 0x02
    assert sent[1].data[2] == 0x32

    assert sent[2].data[0] == 0x23
    assert sent[2].data[1] == 0x03
    assert sent[2].data[2] == 0x32


def test_set_fixed_speed_lighting(lightingNodeProDevice):
    response = '00000000000000000000000000000000'
    lightingNodeProDevice.device.preload_read(Report(0, bytes.fromhex(response)))

    with pytest.raises(NotSupportedByDevice):
        lightingNodeProDevice.set_fixed_speed('sync', 50)

    # check the commands sent
    sent = lightingNodeProDevice.device.sent
    assert len(sent) == 0


def test_set_speed_profile_valid_multi_fan(commanderProDevice):
    ignore = Report(0, bytes(16))
    for _ in range(5):
        commanderProDevice.device.preload_read(ignore)

    commanderProDevice._data.store('temp_sensors_connected', [0x01, 0x01, 0x00, 0x01])
    commanderProDevice._data.store('fan_modes', [0x01, 0x00, 0x01, 0x01, 0x00, 0x00])
    commanderProDevice.set_speed_profile('sync', [(10, 500), (20, 1000)])

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 3

    assert sent[0].data[0] == 0x25
    assert sent[0].data[1] == 0x00
    assert sent[0].data[2] == 0x00

    assert sent[0].data[3] == 0x03
    assert sent[0].data[4] == 0xe8
    assert sent[0].data[15] == 0x01
    assert sent[0].data[16] == 0xf4

    assert sent[1].data[0] == 0x25
    assert sent[1].data[1] == 0x02
    assert sent[1].data[2] == 0x00

    assert sent[2].data[0] == 0x25
    assert sent[2].data[1] == 0x03
    assert sent[2].data[2] == 0x00


def test_set_speed_profile_invalid_temp_sensor(commanderProDevice):
    ignore = Report(0, bytes(16))
    for _ in range(5):
        commanderProDevice.device.preload_read(ignore)

    commanderProDevice._data.store('temp_sensors_connected', [0x01, 0x01, 0x00, 0x01])
    commanderProDevice._data.store('fan_modes', [0x01, 0x00, 0x01, 0x01, 0x00, 0x00])

    commanderProDevice.set_speed_profile('fan1', [(10, 500), (20, 1000)], temperature_sensor=10)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 1

    assert sent[0].data[0] == 0x25
    assert sent[0].data[1] == 0x00
    assert sent[0].data[2] == 0x03

    assert sent[0].data[3] == 0x03
    assert sent[0].data[4] == 0xe8
    assert sent[0].data[15] == 0x01
    assert sent[0].data[16] == 0xf4


def test_set_speed_profile_no_temp_sensors(commanderProDevice):
    ignore = Report(0, bytes(16))
    for _ in range(5):
        commanderProDevice.device.preload_read(ignore)

    commanderProDevice._data.store('temp_sensors_connected', [0x00, 0x00, 0x00, 0x00])
    commanderProDevice._data.store('fan_modes', [0x01, 0x00, 0x01, 0x01, 0x00, 0x00])

    with pytest.raises(ValueError):
        commanderProDevice.set_speed_profile('sync', [(10, 500), (20, 1000)], temperature_sensor=1)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 0


def test_set_speed_profile_valid(commanderProDevice):
    ignore = Report(0, bytes(16))
    for _ in range(5):
        commanderProDevice.device.preload_read(ignore)

    commanderProDevice._data.store('temp_sensors_connected', [0x01, 0x01, 0x00, 0x01])
    commanderProDevice._data.store('fan_modes', [0x01, 0x00, 0x01, 0x01, 0x00, 0x00])
    commanderProDevice.set_speed_profile('fan3', [(10, 500), (20, 1000)])

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 1

    assert sent[0].data[0] == 0x25
    assert sent[0].data[1] == 0x02
    assert sent[0].data[2] == 0x00

    assert sent[0].data[3] == 0x03
    assert sent[0].data[4] == 0xe8
    assert sent[0].data[15] == 0x01
    assert sent[0].data[16] == 0xf4


def test_set_speed_profile_node(lightingNodeProDevice):
    ignore = Report(0, bytes(16))
    for _ in range(5):
        lightingNodeProDevice.device.preload_read(ignore)

    lightingNodeProDevice._data.store('temp_sensors_connected', [0x01, 0x00, 0x00, 0x00])
    lightingNodeProDevice._data.store('fan_modes', [0x01, 0x00, 0x01, 0x01, 0x00, 0x00])

    with pytest.raises(NotSupportedByDevice):
        lightingNodeProDevice.set_speed_profile('sync', [(10, 500), (20, 1000)])

    # check the commands sent
    sent = lightingNodeProDevice.device.sent
    assert len(sent) == 0


def test_set_speed_profile_core(lightingNodeCoreDevice):
    ignore = Report(0, bytes(16))
    for _ in range(5):
        lightingNodeCoreDevice.device.preload_read(ignore)

    lightingNodeCoreDevice._data.store('temp_sensors_connected', [0x01, 0x00, 0x00, 0x00])
    lightingNodeCoreDevice._data.store('fan_modes', [0x01, 0x00, 0x01, 0x01, 0x00, 0x00])

    with pytest.raises(NotSupportedByDevice):
        lightingNodeCoreDevice.set_speed_profile('sync', [(10, 500), (20, 1000)])

    # check the commands sent
    sent = lightingNodeCoreDevice.device.sent
    assert len(sent) == 0


def test_set_speed_profile_valid_unconfigured(commanderProDevice):
    ignore = Report(0, bytes(16))
    for _ in range(5):
        commanderProDevice.device.preload_read(ignore)

    commanderProDevice._data.store('temp_sensors_connected', [0x00, 0x00, 0x00, 0x00])
    commanderProDevice._data.store('fan_modes', [0x01, 0x00, 0x01, 0x01, 0x00, 0x00])

    with pytest.raises(ValueError):
        commanderProDevice.set_speed_profile('fan2', [(10, 500), (20, 1000)])

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 0


def test_set_color_hardware_clear(commanderProDevice):
    ignore = Report(0, bytes(16))
    for _ in range(6):
        commanderProDevice.device.preload_read(ignore)

    effect = {
            'channel': 0x01,
            'start_led': 0x00,
            'num_leds': 0x0f,
            'mode': 0x0a,
            'speed': 0x00,
            'direction': 0x00,
            'random_colors': 0x00,
            'colors': []
        }
    commanderProDevice._data.store('saved_effects', [effect])

    commanderProDevice.set_color('led1', 'clear', [], )

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 0

    effects = commanderProDevice._data.load('saved_effects', default=None)

    assert effects is None


def test_set_color_hardware_off(commanderProDevice):
    ignore = Report(0, bytes(16))
    for _ in range(6):
        commanderProDevice.device.preload_read(ignore)

    effect = {
            'channel': 0x01,
            'start_led': 0x00,
            'num_leds': 0x0f,
            'mode': 0x0a,
            'speed': 0x00,
            'direction': 0x00,
            'random_colors': 0x00,
            'colors': []
        }
    commanderProDevice._data.store('saved_effects', [effect])

    commanderProDevice.set_color('led1', 'off', [])

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 5

    assert sent[3].data[0] == 0x35
    assert sent[3].data[4] == 0x04

    effects = commanderProDevice._data.load('saved_effects', default=None)

    assert effects is None


@pytest.mark.parametrize('directionStr,expected',
                         [('forward', 0x01), ('backward', 0x00)])
def test_set_color_hardware_direction(commanderProDevice, directionStr, expected):
    ignore = Report(0, bytes(16))
    for _ in range(6):
        commanderProDevice.device.preload_read(ignore)

    colors = [[0xaa, 0xbb, 0xcc]]
    commanderProDevice.set_color('led1', 'fixed', colors, direction=directionStr)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 5

    assert sent[3].data[0] == 0x35
    assert sent[3].data[6] == expected

    effects = commanderProDevice._data.load('saved_effects', default=None)

    assert effects is not None
    assert len(effects) == 1


def test_set_color_hardware_direction_default(commanderProDevice):
    ignore = Report(0, bytes(16))
    for _ in range(6):
        commanderProDevice.device.preload_read(ignore)

    colors = [[0xaa, 0xbb, 0xcc]]
    commanderProDevice.set_color('led1', 'fixed', colors)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 5

    assert sent[3].data[0] == 0x35
    assert sent[3].data[6] == 0x01

    effects = commanderProDevice._data.load('saved_effects', default=None)

    assert effects is not None
    assert len(effects) == 1


def test_set_color_hardware_speed_default(commanderProDevice):
    ignore = Report(0, bytes(16))
    for _ in range(6):
        commanderProDevice.device.preload_read(ignore)

    colors = [[0xaa, 0xbb, 0xcc]]
    commanderProDevice.set_color('led1', 'fixed', colors)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 5

    assert sent[3].data[0] == 0x35
    assert sent[3].data[5] == 0x01

    effects = commanderProDevice._data.load('saved_effects', default=None)

    assert effects is not None
    assert len(effects) == 1


@pytest.mark.parametrize('speedStr,expected',
                         [('slow', 0x02), ('fast', 0x00), ('medium', 0x01)])
def test_set_color_hardware_speed(commanderProDevice, speedStr, expected):
    ignore = Report(0, bytes(16))
    for _ in range(6):
        commanderProDevice.device.preload_read(ignore)

    colors = [[0xaa, 0xbb, 0xcc]]
    commanderProDevice.set_color('led1', 'fixed', colors, speed=speedStr)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 5

    assert sent[3].data[0] == 0x35
    assert sent[3].data[5] == expected

    effects = commanderProDevice._data.load('saved_effects', default=None)

    assert effects is not None
    assert len(effects) == 1


def test_set_color_hardware_default_start_end(commanderProDevice):
    ignore = Report(0, bytes(16))
    for _ in range(6):
        commanderProDevice.device.preload_read(ignore)

    colors = [[0xaa, 0xbb, 0xcc]]
    commanderProDevice.set_color('led1', 'fixed', colors)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 5

    assert sent[3].data[0] == 0x35
    assert sent[3].data[2] == 0  # start led
    assert sent[3].data[3] == 204  # num leds

    effects = commanderProDevice._data.load('saved_effects', default=None)

    assert effects is not None
    assert len(effects) == 1


@pytest.mark.parametrize('startLED,expected', [
    (1, 0x00), (30, 0x1d), (92, 0x5b)
    ])
def test_set_color_hardware_start_set(commanderProDevice, startLED, expected):
    ignore = Report(0, bytes(16))
    for _ in range(6):
        commanderProDevice.device.preload_read(ignore)

    colors = [[0xaa, 0xbb, 0xcc]]
    commanderProDevice.set_color('led1', 'fixed', colors, start_led=startLED)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 5

    assert sent[3].data[0] == 0x35
    assert sent[3].data[2] == expected  # start led

    effects = commanderProDevice._data.load('saved_effects', default=None)

    assert effects is not None
    assert len(effects) == 1


@pytest.mark.parametrize('numLED,expected', [
    (1, 1), (30, 30), (96, 96), (203, 203), (204, 204), (205, 204)    ])
def test_set_color_hardware_num_leds(commanderProDevice, numLED, expected):
    ignore = Report(0, bytes(16))
    for _ in range(6):
        commanderProDevice.device.preload_read(ignore)

    colors = [[0xaa, 0xbb, 0xcc]]
    commanderProDevice.set_color('led1', 'fixed', colors, start_led=1, maximum_leds=numLED)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 5

    assert sent[3].data[0] == 0x35
    assert sent[3].data[3] == expected  # num leds

    effects = commanderProDevice._data.load('saved_effects', default=None)

    assert effects is not None
    assert len(effects) == 1

def test_set_color_hardware_too_many_leds(commanderProDevice):
    ignore = Report(0, bytes(16))
    for _ in range(6):
        commanderProDevice.device.preload_read(ignore)

    colors = [[0xaa, 0xbb, 0xcc]]
    commanderProDevice.set_color('led1', 'fixed', colors, start_led=200, maximum_leds=50)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 5

    assert sent[3].data[0] == 0x35
    assert sent[3].data[2] == 0xc7  # start led
    assert sent[3].data[3] == 5  # num led

    effects = commanderProDevice._data.load('saved_effects', default=None)

    assert effects is not None
    assert len(effects) == 1


def test_set_color_hardware_too_few_leds(commanderProDevice):
    ignore = Report(0, bytes(16))
    for _ in range(6):
        commanderProDevice.device.preload_read(ignore)

    colors = [[0xaa, 0xbb, 0xcc]]
    commanderProDevice.set_color('led1', 'fixed', colors, start_led=1, maximum_leds=0)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 5

    assert sent[3].data[0] == 0x35
    assert sent[3].data[2] == 0x00  # start led
    assert sent[3].data[3] == 0x01  # num led

    effects = commanderProDevice._data.load('saved_effects', default=None)

    assert effects is not None
    assert len(effects) == 1


@pytest.mark.parametrize('channel,expected',
                         [('led1', 0x00), ('led2', 0x01)])
def test_set_color_hardware_channel(commanderProDevice, channel, expected):
    ignore = Report(0, bytes(16))
    for _ in range(6):
        commanderProDevice.device.preload_read(ignore)

    colors = [[0xaa, 0xbb, 0xcc]]
    commanderProDevice.set_color(channel, 'fixed', colors)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 5

    assert sent[0].data[1] == expected
    assert sent[1].data[1] == expected
    assert sent[2].data[1] == expected
    assert sent[3].data[0] == 0x35
    assert sent[3].data[1] == expected

    effects = commanderProDevice._data.load('saved_effects', default=None)

    assert effects is not None
    assert len(effects) == 1


def test_set_color_hardware_sync_channel(commanderProDevice):
    ignore = Report(0, bytes(16))
    for _ in range(9):
        commanderProDevice.device.preload_read(ignore)

    colors = [[0xaa, 0xbb, 0xcc]]
    commanderProDevice.set_color('sync', 'fixed', colors)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 2*(3 + 1) + 1

    assert sent[0].data[1] == 0
    assert sent[1].data[1] == 0
    assert sent[2].data[1] == 0

    assert sent[3].data[1] == 1
    assert sent[4].data[1] == 1
    assert sent[5].data[1] == 1

    assert sent[6].data[0] == 0x35
    assert sent[6].data[1] == 0

    assert sent[7].data[0] == 0x35
    assert sent[7].data[1] == 1

    effects = commanderProDevice._data.load('saved_effects', default=None)

    assert effects is not None
    assert len(effects) == 2


def test_set_color_hardware_random_color(commanderProDevice):
    ignore = Report(0, bytes(16))
    for _ in range(6):
        commanderProDevice.device.preload_read(ignore)

    colors = []
    commanderProDevice.set_color('led1', 'fixed', colors)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 5

    assert sent[3].data[0] == 0x35
    assert sent[3].data[7] == 0x01

    effects = commanderProDevice._data.load('saved_effects', default=None)

    assert effects is not None
    assert len(effects) == 1


def test_set_color_hardware_not_random_color(commanderProDevice):
    ignore = Report(0, bytes(16))
    for _ in range(6):
        commanderProDevice.device.preload_read(ignore)

    colors = [[0xaa, 0xbb, 0xcc]]
    commanderProDevice.set_color('led1', 'fixed', colors)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 5

    assert sent[3].data[0] == 0x35
    assert sent[3].data[7] == 0x00

    effects = commanderProDevice._data.load('saved_effects', default=None)

    assert effects is not None
    assert len(effects) == 1


def test_set_color_hardware_too_many_colors(commanderProDevice):
    ignore = Report(0, bytes(16))
    for _ in range(6):
        commanderProDevice.device.preload_read(ignore)

    colors = [[0xaa, 0xbb, 0xcc], [0x00, 0x11, 0x22], [0x33, 0x44, 0x55]]
    commanderProDevice.set_color('led1', 'fixed', colors)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 5

    assert sent[3].data[0] == 0x35
    assert sent[3].data[7] == 0x00

    assert sent[3].data[9] == 0xaa
    assert sent[3].data[10] == 0xbb
    assert sent[3].data[11] == 0xcc

    effects = commanderProDevice._data.load('saved_effects', default=None)

    assert effects is not None
    assert len(effects) == 1


def test_set_color_hardware_too_few_colors(commanderProDevice):
    ignore = Report(0, bytes(16))
    for _ in range(6):
        commanderProDevice.device.preload_read(ignore)

    commanderProDevice.set_color('led1', 'fixed', [])

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 5

    assert sent[3].data[0] == 0x35
    assert sent[3].data[7] == 0x01

    assert sent[3].data[9] == 0x00
    assert sent[3].data[10] == 0x00
    assert sent[3].data[11] == 0x00

    effects = commanderProDevice._data.load('saved_effects', default=None)

    assert effects is not None
    assert len(effects) == 1


@pytest.mark.parametrize('modeStr,expected', [
    ('rainbow', 0x00), ('color_shift', 0x01), ('color_pulse', 0x02),
    ('color_wave', 0x03), ('fixed', 0x04), ('visor', 0x06), ('marquee', 0x07),
    ('blink', 0x08), ('sequential', 0x09), ('rainbow2', 0x0a),
    ])
def test_set_color_hardware_valid_mode(commanderProDevice, modeStr, expected):
    ignore = Report(0, bytes(16))
    for _ in range(6):
        commanderProDevice.device.preload_read(ignore)

    commanderProDevice.set_color('led1', modeStr, [])

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 5

    assert sent[3].data[0] == 0x35
    assert sent[3].data[4] == expected

    effects = commanderProDevice._data.load('saved_effects', default=None)

    assert effects is not None
    assert len(effects) == 1


def test_set_color_hardware_invalid_mode(commanderProDevice):
    ignore = Report(0, bytes(16))
    for _ in range(6):
        commanderProDevice.device.preload_read(ignore)

    colors = [[0xaa, 0xbb, 0xcc]]

    with pytest.raises(ValueError):
        commanderProDevice.set_color('led1', 'invalid', colors)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 0

    effects = commanderProDevice._data.load('saved_effects', default=None)
    assert effects is None


def test_set_color_hardware_multipe_commands(commanderProDevice):
    ignore = Report(0, bytes(16))
    for _ in range(6):
        commanderProDevice.device.preload_read(ignore)

    effect = {
            'channel': 0x01,
            'start_led': 0x00,
            'num_leds': 0x0f,
            'mode': 0x0a,
            'speed': 0x00,
            'direction': 0x00,
            'random_colors': 0x00,
            'colors': [0xaa, 0xbb, 0xcc]
        }
    commanderProDevice._data.store('saved_effects', [effect])

    commanderProDevice.set_color('led1', 'fixed', [[0x00, 0x11, 0x22]], start_led=16, maximum_leds=5)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 6

    assert sent[3].data[0] == 0x35
    assert sent[4].data[0] == 0x35

    effects = commanderProDevice._data.load('saved_effects', default=None)

    assert effects is not None
    assert len(effects) == 2


def test_set_color_hardware_last_commands(commanderProDevice):

    for i in range(15):
        commanderProDevice.device.preload_read(Report(0, bytes.fromhex('00000000000000000000000000000000')))

    effect = {
            'channel': 0x01,
            'start_led': 0x00,
            'num_leds': 0x0f,
            'mode': 0x0a,
            'speed': 0x00,
            'direction': 0x00,
            'random_colors': 0x00,
            'colors': [0xaa, 0xbb, 0xcc]
        }
    commanderProDevice._data.store('saved_effects', [effect, effect, effect, effect, effect, effect, effect])

    commanderProDevice.set_color('led1', 'fixed', [[0x00, 0x11, 0x22]], start_led=16, maximum_leds=5)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 12

    effects = commanderProDevice._data.load('saved_effects', default=None)

    assert effects is not None
    assert len(effects) == 8



def test_set_color_hardware_max_commands(commanderProDevice):

    for i in range(15):
        commanderProDevice.device.preload_read(Report(0, bytes.fromhex('00000000000000000000000000000000')))

    effect = {
            'channel': 0x01,
            'start_led': 0x00,
            'num_leds': 0x0f,
            'mode': 0x0a,
            'speed': 0x00,
            'direction': 0x00,
            'random_colors': 0x00,
            'colors': [0xaa, 0xbb, 0xcc]
        }
    commanderProDevice._data.store('saved_effects', [effect, effect, effect, effect, effect, effect, effect, effect])

    commanderProDevice.set_color('led1', 'fixed', [[0x00, 0x11, 0x22]], start_led=16, maximum_leds=5)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 0

    effects = commanderProDevice._data.load('saved_effects', default=None)

    assert effects is not None
    assert len(effects) == 8


def test_set_color_hardware_too_many_commands(commanderProDevice):
    ignore = Report(0, bytes(16))
    for _ in range(6):
        commanderProDevice.device.preload_read(ignore)

    effect = {
            'channel': 0x01,
            'start_led': 0x00,
            'num_leds': 0x0f,
            'mode': 0x0a,
            'speed': 0x00,
            'direction': 0x00,
            'random_colors': 0x00,
            'colors': [0xaa, 0xbb, 0xcc]
        }
    commanderProDevice._data.store('saved_effects', [effect, effect, effect, effect, effect, effect, effect, effect, effect])

    commanderProDevice.set_color('led1', 'fixed', [[0x00, 0x11, 0x22]], start_led=16, maximum_leds=5)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 0


def test_send_command_valid_data(commanderProDevice):
    ignore = Report(0, bytes(16))
    for _ in range(2):
        commanderProDevice.device.preload_read(ignore)

    commanderProDevice._send_command(6, [255, 0, 20, 10, 15])

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 1
    assert len(sent[0].data) == 64
    assert sent[0].data[0] == 6
    assert sent[0].data[1] == 255


def test_send_command_no_data(commanderProDevice):
    ignore = Report(0, bytes(16))
    for _ in range(2):
        commanderProDevice.device.preload_read(ignore)

    commanderProDevice._send_command(6)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 1
    assert len(sent[0].data) == 64
    assert sent[0].data[0] == 6
    assert sent[0].data[1] == 0


def test_send_command_data_too_long(commanderProDevice):
    ignore = Report(0, bytes(16))
    for _ in range(2):
        commanderProDevice.device.preload_read(ignore)

    data = bytearray(100)
    commanderProDevice._send_command(3, data)

    # check the commands sent
    sent = commanderProDevice.device.sent
    assert len(sent) == 1
    assert len(sent[0].data) == 64
    assert sent[0].data[0] == 3
    assert sent[0].data[1] == 0



"""liquidctl drivers for Corsair HXi and RMi series power supply units.

Supported devices:

- Corsair HXi (HX750i, HX850i, HX1000i and HX1200i)
- Corsair RMi (RM650i, RM750i, RM850i and RM1000i)

Copyright Jonas Malaco and contributors

Port of corsaiRMi by notaz and realies.
Copyright (c) notaz, 2016

Incorporates or uses as reference work by Sean Nelson.

SPDX-License-Identifier: GPL-3.0-or-later
"""

import logging
from datetime import timedelta
from enum import Enum

from liquidctl.driver.usb import UsbHidDriver
from liquidctl.error import NotSupportedByDevice
from liquidctl.pmbus import CommandCode as CMD
from liquidctl.pmbus import WriteBit, linear_to_float
from liquidctl.util import clamp

_LOGGER = logging.getLogger(__name__)

_REPORT_LENGTH = 64
_SLAVE_ADDRESS = 0x02
_CORSAIR_READ_TOTAL_UPTIME = CMD.MFR_SPECIFIC_D1
_CORSAIR_READ_UPTIME = CMD.MFR_SPECIFIC_D2
_CORSAIR_12V_OCP_MODE = CMD.MFR_SPECIFIC_D8
_CORSAIR_READ_OUTPUT_POWER = CMD.MFR_SPECIFIC_EE
_CORSAIR_FAN_CONTROL_MODE = CMD.MFR_SPECIFIC_F0

_RAIL_12V = 0x0
_RAIL_5V = 0x1
_RAIL_3P3V = 0x2
_RAIL_NAMES = {_RAIL_12V: '+12V', _RAIL_5V: '+5V', _RAIL_3P3V: '+3.3V'}
_MIN_FAN_DUTY = 0


class OCPMode(Enum):
    """Overcurrent protection mode."""

    SINGLE_RAIL = 0x1
    MULTI_RAIL = 0x2

    def __str__(self):
        return self.name.capitalize().replace('_', ' ')


class FanControlMode(Enum):
    """Fan control mode."""

    HARDWARE = 0x0
    SOFTWARE = 0x1

    def __str__(self):
        return self.name.capitalize()


class CorsairHidPsu(UsbHidDriver):
    """Corsair HXi or RMi series power supply unit."""

    # support for hwmon: corsair-psu, Linux 5.11 (5.13 recommended)

    _MATCHES = [
        (0x1b1c, 0x1c05, 'Corsair HX750i', {
            'fpowin115': (0.00013153276902318052, 1.0118732314945875, 9.783796618886313),
            'fpowin230': ( 9.268856467314546e-05, 1.0183515407387007, 8.279822175342481),
        }),
        (0x1b1c, 0x1c06, 'Corsair HX850i', {
            'fpowin115': (0.00011552923724840388, 1.0111311876704099, 12.015296651918918),
            'fpowin230': ( 8.126644224872423e-05, 1.0176256272095185, 10.290640442373850),
        }),
        (0x1b1c, 0x1c07, 'Corsair HX1000i', {  # CP-9020074-NA
            'fpowin115': (9.48609754417109e-05,  1.0170509865269720, 11.619826520447452),
            'fpowin230': (9.649987544008507e-05, 1.0018241767296636, 12.759957859756842),
        }),
        (0x1b1c, 0x1c08, 'Corsair HX1200i', {
            'fpowin115': (6.244705156199815e-05,  1.0234738310580973, 15.293509559389241),
            'fpowin230': (5.9413179794350966e-05, 1.0023670927127724, 15.886126793547152),
        }),
        (0x1b1c, 0x1c0a, 'Corsair RM650i', {
            'fpowin115': (0.00017323493381072683, 1.0047044721686030, 12.376592422281606),
            'fpowin230': (0.00012413136310310370, 1.0284317478987164,  9.465259079360674),
        }),
        (0x1b1c, 0x1c0b, 'Corsair RM750i', {
            'fpowin115': (0.00015013694263596336, 1.0047044721686027, 14.280683564171110),
            'fpowin230': (0.00010460621468919797, 1.0173089573727216, 11.495900706372142),
        }),
        (0x1b1c, 0x1c0c, 'Corsair RM850i', {
            'fpowin115': (0.00012280002467981107, 1.0159421430340847, 13.555472968718759),
            'fpowin230': ( 8.816054254801031e-05, 1.0234738318592156, 10.832902491655597),
        }),
        (0x1b1c, 0x1c0d, 'Corsair RM1000i', {
            'fpowin115': (0.00010018433053123574, 1.0272313660072225, 14.092187353321624),
            'fpowin230': ( 8.600634771656125e-05, 1.0289245073649413, 13.701515390258626),
        }),
        (0x1b1c, 0x1c1e, 'Corsair HX1000i (2022) (experimental)', {  # CP-9020214-NA
            # FIXME does not accurately represent known efficiency curve at 2%–10% loads
            'fpowin115': (0.12038623471806167, 0.9899868098825505, 0.01312560158286225),
            'fpowin230': (0.08725695213088608, 1.0017598020558338, 0.009789546119565251),
        }),
        (0x1b1c, 0x1c1f, 'Corsair HX1500i (experimental)', {  # CP-9020215-NA
            # FIXME does not accurately represent known efficiency curve at 2%–10% loads
            'fpowin115': (0.099089523528618,   1.0125991459969408, 0.011978189078640221),
            'fpowin230': (0.06951642350659733, 1.0183515407262729, 0.011039762908188666),
        }),
    ]

    def __init__(self, *args, fpowin115=None, fpowin230=None, **kwargs):
        assert fpowin115 and fpowin230

        super().__init__(*args, **kwargs)
        self.fpowin115 = fpowin115
        self.fpowin230 = fpowin230

    def initialize(self, single_12v_ocp=False, direct_access=False, **kwargs):
        """Initialize the device and the driver.

        This method should be called every time the systems boots, resumes from
        a suspended state, or if the device has just been (re)connected.  In
        those scenarios, no other method, except `connect()` or `disconnect()`,
        should be called until the device and driver has been (re-)initialized.

        Returns None or a list of `(property, value, unit)` tuples, similarly
        to `get_status()`.
        """

        # necessary to receive non-zero status responses from the device:
        # replies before calling this function appear to follow the pattern
        # <address> <cte 0xfe> <zero> <zero> <padding...>

        if self._hwmon:
            if not direct_access:
                _LOGGER.warning('bound to %s kernel driver, OCP and fan modes not changed',
                                self._hwmon.driver)
                return
            else:
                _LOGGER.warning('forcing re-initialization despite %s kernel driver',
                                self._hwmon.driver)

        self._write([0xfe, 0x03])
        _ = self._read()

        # don't check current OCP and fan control modes in case we're racing
        # with a hwmon driver and the returned values, which aren't available
        # through hwmon, are corrupted by the race

        mode = OCPMode.SINGLE_RAIL if single_12v_ocp else OCPMode.MULTI_RAIL
        _LOGGER.info('setting +12V OCP mode to %s', mode)
        self._exec(WriteBit.WRITE, _CORSAIR_12V_OCP_MODE, [mode.value])

        _LOGGER.info('resetting fan control to hardware mode')
        self._set_fan_control_mode(FanControlMode.HARDWARE)

    def _get_status_directly(self):
        self.device.clear_enqueued_reports()
        ret = self._exec(WriteBit.WRITE, CMD.PAGE, [0])
        if ret[1] == 0xfe:
            _LOGGER.warning('possibly uninitialized device')

        input_voltage = self._get_float(CMD.READ_VIN)

        ret = [
            ('Current uptime', self._get_timedelta(_CORSAIR_READ_UPTIME), ''),
            ('Total uptime', self._get_timedelta(_CORSAIR_READ_TOTAL_UPTIME), ''),
            ('VRM temperature', self._get_float(CMD.READ_TEMPERATURE_1), '°C'),
            ('Case temperature', self._get_float(CMD.READ_TEMPERATURE_2), '°C'),
            ('Fan control mode', self._get_fan_control_mode(), ''),
            ('Fan speed', self._get_float(CMD.READ_FAN_SPEED_1), 'rpm'),
            ('Input voltage', input_voltage, 'V'),
            ('+12V OCP mode', self._get_12v_ocp_mode(), ''),
        ]

        for rail in [_RAIL_12V, _RAIL_5V, _RAIL_3P3V]:
            name = _RAIL_NAMES[rail]
            self._exec(WriteBit.WRITE, CMD.PAGE, [rail])
            ret.append((f'{name} output voltage', self._get_float(CMD.READ_VOUT), 'V'))
            ret.append((f'{name} output current', self._get_float(CMD.READ_IOUT), 'A'))
            ret.append((f'{name} output power', self._get_float(CMD.READ_POUT), 'W'))

        output_power = self._get_float(_CORSAIR_READ_OUTPUT_POWER)
        input_power = round(self._input_power_at(input_voltage, output_power), 0)
        efficiency = round(output_power / input_power * 100, 0)

        ret.append(('Total power output', output_power, 'W'))
        ret.append(('Estimated input power', input_power, 'W'))
        ret.append(('Estimated efficiency', efficiency, '%'))

        self._exec(WriteBit.WRITE, CMD.PAGE, [0])
        return ret

    def _get_status_from_hwmon(self):
        # can't report some values (current and total uptime are only available
        # on debugfs, and fan and ocp modes are not available at all); still,
        # with this particular device, it is better to ignore them than to race
        # with a kernel driver
        _LOGGER.warning('some attributes cannot be read from %s kernel driver', self._hwmon.driver)

        input_voltage = self._hwmon.read_int('in0_input') * 1e-3

        ret = [
            ('VRM temperature', self._hwmon.read_int('temp1_input') * 1e-3, '°C'),
            ('Case temperature', self._hwmon.read_int('temp2_input') * 1e-3, '°C'),
            ('Fan speed', self._hwmon.read_int('fan1_input'), 'rpm'),
            ('Input voltage', input_voltage, 'V'),
        ]

        for n, rail in zip(range(2, 5), [_RAIL_12V, _RAIL_5V, _RAIL_3P3V]):
            i = n - 1
            name = _RAIL_NAMES[rail]
            ret.append((f'{name} output voltage', self._hwmon.read_int(f'in{i}_input') * 1e-3, 'V'))
            ret.append((f'{name} output current', self._hwmon.read_int(f'curr{n}_input') * 1e-3, 'A'))
            ret.append((f'{name} output power', self._hwmon.read_int(f'power{n}_input') * 1e-6, 'W'))

        output_power = self._hwmon.read_int('power1_input') * 1e-6
        input_power = round(self._input_power_at(input_voltage, output_power), 0)
        efficiency = round(output_power / input_power * 100, 0)

        ret.append(('Total power output', output_power, 'W'))
        ret.append(('Estimated input power', input_power, 'W'))
        ret.append(('Estimated efficiency', efficiency, '%'))

        return ret

    def get_status(self, direct_access=False, **kwargs):
        """Get a status report.

        Returns a list of `(property, value, unit)` tuples.
        """

        if self._hwmon and not direct_access:
            _LOGGER.info('bound to %s kernel driver, reading status from hwmon', self._hwmon.driver)
            return self._get_status_from_hwmon()

        if self._hwmon:
            _LOGGER.warning('directly reading the status despite %s kernel driver',
                            self._hwmon.driver)

        return self._get_status_directly()

    def set_fixed_speed(self, channel, duty, **kwargs):
        """Set channel to a fixed speed duty."""
        duty = clamp(duty, _MIN_FAN_DUTY, 100)
        _LOGGER.info('ensuring fan control is in software mode')
        self._set_fan_control_mode(FanControlMode.SOFTWARE)
        _LOGGER.info('setting fan PWM duty to %d%%', duty)
        self._exec(WriteBit.WRITE, CMD.FAN_COMMAND_1, [duty])

    def set_color(self, channel, mode, colors, **kwargs):
        """Not supported by this device."""
        raise NotSupportedByDevice()

    def set_speed_profile(self, channel, profile, **kwargs):
        """Not supported by this device."""
        raise NotSupportedByDevice()

    def _input_power_at(self, input_voltage, output_power):
        def quadratic(params, x):
            a, b, c = params
            return a * x**2 + b * x + c

        for_in115v = quadratic(self.fpowin115, output_power)
        for_in230v = quadratic(self.fpowin230, output_power)
        _LOGGER.debug('input power estimates: %.3f W @ 115 V; %.3f W @ 230 V', for_in115v, for_in230v)

        # linearly interpolate for input_voltage
        return for_in115v + (for_in230v - for_in115v) / 115 * (input_voltage - 115)

    def _write(self, data):
        assert len(data) <= _REPORT_LENGTH
        packet = bytearray(1 + _REPORT_LENGTH)
        packet[1: 1 + len(data)] = data  # device doesn't use numbered reports
        self.device.write(packet)

    def _read(self):
        return self.device.read(_REPORT_LENGTH)

    def _exec(self, writebit, command, data=None):
        out = [_SLAVE_ADDRESS | WriteBit(writebit), CMD(command)] + (data or [])
        self._write(out)
        ret = self._read()
        assert ret[0:2] == out[0:2], f'invalid response (possible conflict with another program)'
        return ret

    def _get_12v_ocp_mode(self):
        """Get +12V single/multi-rail OCP mode."""
        return OCPMode(self._exec(WriteBit.READ, _CORSAIR_12V_OCP_MODE)[2])

    def _get_fan_control_mode(self):
        """Get hardware/software fan control mode."""
        return FanControlMode(self._exec(WriteBit.READ, _CORSAIR_FAN_CONTROL_MODE)[2])

    def _set_fan_control_mode(self, mode):
        """Set hardware/software fan control mode."""
        return self._exec(WriteBit.WRITE, _CORSAIR_FAN_CONTROL_MODE, [mode.value])

    def _get_float(self, command):
        """Get float value with `command`."""
        return linear_to_float(self._exec(WriteBit.READ, command)[2:])

    def _get_timedelta(self, command):
        """Get timedelta with `command`."""
        secs = int.from_bytes(self._exec(WriteBit.READ, command)[2:], byteorder='little')
        return timedelta(seconds=secs)

    def set_screen(self, channel, mode, value, **kwargs):
        """Not supported by this device."""
        raise NotSupportedByDevice()


# deprecated aliases
CorsairHidPsuDriver = CorsairHidPsu



import pytest
from _testutils import MockHidapiDevice, Report

from liquidctl.driver.hwmon import HwmonDevice
from liquidctl.driver.aquacomputer import Aquacomputer
from liquidctl.error import NotSupportedByDriver, NotSupportedByDevice

D5NEXT_SAMPLE_STATUS_REPORT = bytes.fromhex(
    "00030DCB597C00010000006403FF00000051000004DC14000001E0007A98AF000"
    "00000FFFF000041A803C169000001481ACAA3465CB804B401F40000005213887F"
    "FF7FFF7FFF7FFF7FFF7FFF7FFF000000000000000009D27FFF00007FFF01F404B"
    "400200026016D006300000004B200D7010207B80000000000098D083A098A083A"
    "00060001000000000000000000000000011A24015E27101D4CFFBF"
)

D5NEXT_SAMPLE_CONTROL_REPORT = bytes.fromhex(
    "00031E00000000000AC0007FFF0000000002020E100BB8000000000A0001000A0"
    "006000A000C000A0000000000000101F42710271007D000000027102710138802"
    "07D200000C8001F4012C00000064001E00010AF00A8C0AFD0B4C0B9D0BE90C460"
    "C9F0CF30D3C0DA20DE50E420E8A0EE60F350F7000000000000002D604D606D609"
    "810A010DAC1202162D17AD19D81EAE222E232E0212D300000D4801F4012C00000"
    "064001E00010AF00A8C0AFA0B4C0BA40C000C4F0CA30D110D510DA60DFD0E560E"
    "9E0EEE0F2010820000008C0000000000000000000001000180035407810A810B0"
    "10C810DD70EAC03E8FF000000000F030000FFFF0F19000003E80164000003E801"
    "FF0032006400000000000000000000000000000000000000000000FFFF0000FFF"
    "F0000FFFF0000FFFF0000FFFF0000FFFF000F0F080000FFFF0F19000003E80164"
    "000003E801FF00190028001400000000000000000000000000000000000000000"
    "00F03E7FFFF00FEFFFF0000FFFF0000FFFF0000FFFF001E0F0B0000FFFF0F1900"
    "0003E80164000003E801FF001E002800010006005000000000000000000000000"
    "0000002FF02FF01FBFFFF0525FFFF00C5FFFF03F5FFFF05F3FFFF002D0F040006"
    "FFFF0F19000003E80164000003E801FF002800050000000000000000000000000"
    "0000000000000000000000F0000FFFF01FDFFFF03FFFFFF00FAFFFF01CE10FF00"
    "3C0F040006FFFF0F19000003E80164000003E801FF00280005000000000000000"
    "00000000000000000000000000000000F00FAFFFF05DCFFFF01C2FFFF0000FFFF"
    "07D010FF004B0F040006FFFF0F19000003E80164000003E801FF0028000500000"
    "000000000000000000000000000000000000000000F03E8FFFF01C2FFFF0000FF"
    "FF0064FFFF032010FF010006030000FFFF0F19000003E80164000003E801FF001"
    "E006400000000000000000000000000000000000000000000FFFF0000FFFF0000"
    "FFFF0000FFFF0000FFFF0000FFFF010006000000FFFF0F19000003E8016400000"
    "3E80164001E006400000000000000000000000000000000000000000000FFFF00"
    "00FFFF0000FFFF0000FFFF0000FFFF0000FFFFC00401C20FA00110FB"
)

FARBWERK360_SAMPLE_STATUS_REPORT = bytes.fromhex(
    "000141BBDE9203E80000006403FE000000110000001A150000005F0008AE3E000"
    "00023BFC8C01AA20EFFD6A0E8A3915AEC0A3C0A470A6F09F814507FFF7FFF7FFF"
    "7FFF7FFF7FFF7FFF7FFF7FFF7FFF7FFF7FFF7FFF7FFF7FFF00000000000000000"
    "00000000000000001F901FA0006000000030000004300000000000A0324000000"
    "00000000002710271027102710271003E8000003E8000003E8000003E80000000"
    "0000000000000000000010002000101040006"
)

OCTO_SAMPLE_STATUS_REPORT = bytes.fromhex(
    "00023A92C9EA03E80001006503FB000000010000010DB4000000C5003C3EA4010"
    "00200000000000000000000000000059EDCFFDCFFDDFFDDA7A65BF80AC60ACF0B"
    "150D600EC87FFF7FFF7FFF7FFF7FFF7FFF7FFF7FFF7FFF7FFF7FFF7FFF7FFF7FF"
    "F7FFF0300000000000000000000000000000004B9000300030000055D04B90001"
    "00010000000008138804B9015E006702400000000000000000000000000000000"
    "00000000000000000000000000000000000000000000000000000000000000000"
    "0000000000000000000000000000000000000000213B04B900020002000000000"
    "80000000003E8055D0000000003E800000000000003E800000000000003E80000"
    "0000000003E800000000000003E800000000000003E800000000000003E8213B0"
    "000000003E827100000000003E827100000000000000000120412862710271098"
    "20"
)

OCTO_SAMPLE_CONTROL_REPORT = bytes.fromhex(
    "000228000000A90000051402BC000000000001F42710271007D00201F42710271"
    "007D00201F42710271007D00201F42710271007D00201F42710271007D00201F4"
    "2710271007D00201F42710271007D00001F42710271007D000055DFFFF0DAC057"
    "804B000000028001400010AF00A8C0AFA0B4A0BA40BF40C4E0C9D0CF80D480DA2"
    "0DF20E4C0E9C0EF50F460FA00000008C011801F4032004B0069008D40B680E4C1"
    "194152C19281D7422102710000000FFFF0DAC057804B000000028001400010AF0"
    "0A8C0AFA0B4A0BA40BF40C4E0C9D0CF80D480DA20DF20E4C0E9C0EF50F460FA00"
    "000008C011801F4032004B0069008D40B680E4C1194152C19281D742210271000"
    "0000FFFF0DAC057804B000000028001400010AF00A8C0AFA0B4A0BA40BF40C4E0"
    "C9D0CF80D480DA20DF20E4C0E9C0EF50F460FA00000008C011801F4032004B006"
    "9008D40B680E4C1194152C19281D7422102710000000FFFF0DAC057804B000000"
    "028001400010AF00A8C0AFA0B4A0BA40BF40C4E0C9D0CF80D480DA20DF20E4C0E"
    "9C0EF50F460FA00000008C011801F4032004B0069008D40B680E4C1194152C192"
    "81D7422102710000000FFFF0DAC057804B000000028001400010AF00A8C0AFA0B"
    "4A0BA40BF40C4E0C9D0CF80D480DA20DF20E4C0E9C0EF50F460FA00000008C011"
    "801F4032004B0069008D40B680E4C1194152C19281D7422102710000000FFFF0D"
    "AC057804B000000028001400010AF00A8C0AFA0B4A0BA40BF40C4E0C9D0CF80D4"
    "80DA20DF20E4C0E9C0EF50F460FA00000008C011801F4032004B0069008D40B68"
    "0E4C1194152C19281D7422102710000000FFFF0DAC057804B0000000280014000"
    "10AF00A8C0AFA0B4A0BA40BF40C4E0C9D0CF80D480DA20DF20E4C0E9C0EF50F46"
    "0FA00000008C011801F4032004B0069008D40B680E4C1194152C19281D7422102"
    "71000213BFFFF0DAC057804B000000028001400010AF00A8C0AFA0B4A0BA40BF4"
    "0C4E0C9D0CF80D480DA20DF20E4C0E9C0EF50F460FA00000008C011801F403200"
    "4B0069008D40B680E4C1194152C19281D74221027100000FF000000000F030000"
    "FFFF0F19000003E80164000003E801FF003200640000000000000000000000000"
    "0000000000000000000FFFF0000FFFF0000FFFF0000FFFF0000FFFF0000FFFF00"
    "0F0F080000FFFF0F19000003E80164000003E801FF00190028001400000000000"
    "00000000000000000000000000000000F03E7FFFF00FEFFFF0000FFFF0000FFFF"
    "0000FFFF001E0F0B0000FFFF0F19000003E80164000003E801FF001E002800010"
    "0060050000000000000000000000000000002FF02FF01FBFFFF0525FFFF00C5FF"
    "FF03F5FFFF05F3FFFF002D0F130000FFFF0F19000003E80164000003E801FF001"
    "9000A0005000500190000000000000000000000000000000000FF0200FF780000"
    "FFFF0000FFFF0000FFFF0000FFFF003C0F040006FFFF0F19000003E8016400000"
    "3E801FF0028000500000000000000000000000000000000000000000000000F00"
    "00FFFF01FDFFFF03FFFFFF00FAFFFF01CE10FF004B0F0F0000FFFF0F19000003E"
    "80164000003E801FF00280004001E001E00000000000000000000000000000000"
    "00000000007800780000FFFF0000FFFF0000FFFF0000FFFF01000F030000FFFF0"
    "F19000003E80164000003E801FF00320064000000000000000000000000000000"
    "00000000000000FFFF0000FFFF0000FFFF0000FFFF0000FFFF0000FFFF010F0F0"
    "80000FFFF0F19000003E80164000003E801FF0019002800140000000000000000"
    "000000000000000000000000000F03E7FFFF00FEFFFF0000FFFF0000FFFF0000F"
    "FFF011E0F0B0000FFFF0F19000003E80164000003E801FF001E00280001000600"
    "50000000000000000000000000000002FF02FF01FBFFFF0525FFFF00C5FFFF03F"
    "5FFFF05F3FFFF012D0F130000FFFF0F19000003E80164000003E801FF0019000A"
    "0005000500190000000000000000000000000000000000FF0200FF780000FFFF0"
    "000FFFF0000FFFF0000FFFF013C0F040006FFFF0F19000003E80164000003E801"
    "FF0028000500000000000000000000000000000000000000000000000F0000FFF"
    "F01FDFFFF03FFFFFF00FAFFFF01CE10FF014B0F0F0000FFFF0F19000003E80164"
    "000003E801FF00280004001E001E0000000000000000000000000000000000000"
    "000007800780000FFFF0000FFFF0000FFFF0000FFFF0100001388138813881388"
    "015E01AB59"
)

QUADRO_SAMPLE_STATUS_REPORT = bytes.fromhex(
    "00035B72FF4000010000006504080000000100000013C5000000910032CBB0000"
    "0000000000000FFD5FFD69B54FFD8A6FD5B977FFF7FFF06517FFF09597FFF7FFF"
    "7FFF7FFF7FFF7FFF7FFF7FFF7FFF7FFF7FFF13887FFF7FFF7FFF0300000000000"
    "000000000000300000004B9000000000000000000000000000000271004B90000"
    "0000000000000805BB04B900000000016400000015E004B900000000000000000"
    "80000000003E800000000000003E827100000000003E805BB0000000003E815E0"
    "0000000003E82710000A0000000E000000002710FF000001"
)

QUADRO_SAMPLE_CONTROL_REPORT = bytes.fromhex(
    "00031C000000A9000002580514FAEC05DC0001F42710271007D00001F42710271"
    "007D00001F42710271007D00001F42710271007D0000000FFFF0DAC057804B000"
    "000028001400010AF00A8C0AFA0B4A0BA40BF40C4E0C9D0CF80D480DA20DF20E4"
    "C0E9C0EF50F460FA00000008C011801F4032004B0069008D40B680E4C1194152C"
    "19281D7422102710004CD0FFFF0DAC057804B000000028001400010AF00A8C0AF"
    "A0B4A0BA40BF40C4E0C9D0CF80D480DA20DF20E4C0E9C0EF50F460FA00000008C"
    "011801F4032004B0069008D40B680E4C1194152C19281D74221027100005BB000"
    "30DAC057804B000000028001400010AF00A8C0AFA0B4A0BA40BF40C4E0C9D0CF8"
    "0D480DA20DF20E4C0E9C0EF50F460FA00000008C011801F4032004B0069008D40"
    "B680E4C1194152C19281D74221027100015E0FFFF0DAC057804B0000000280014"
    "00010AF00A8C0AFA0B4A0BA40BF40C4E0C9D0CF80D480DA20DF20E4C0E9C0EF50"
    "F460FA00000008C011801F4032004B0069008D40B680E4C1194152C19281D7422"
    "102710FF000200000F030000FFFF0F19000003E80164000003E801FF003200640"
    "0000000000000000000000000000000000000000000FFFF0000FFFF0000FFFF00"
    "00FFFF0000FFFF0000FFFF000F0F080000FFFF0F19000003E80164000003E801F"
    "F0019002800140000000000000000000000000000000000000000000F03E7FFFF"
    "00FEFFFF0000FFFF0000FFFF0000FFFF001E0F0B0000FFFF0F19000003E801640"
    "00003E801FF001E0028000100060050000000000000000000000000000002FF02"
    "FF01FBFFFF0525FFFF00C5FFFF03F5FFFF05F3FFFF002D0F040006FFFF0F19000"
    "003E80164000003E801FF00280005000000000000000000000000000000000000"
    "00000000000F0000FFFF01FDFFFF03FFFFFF00FAFFFF01CE10FF003C0F040006F"
    "FFF0F19000003E80164000003E801FF0028000200000000000000000000000000"
    "000000000000000000000F03FFFFFF07D0FFFF0000FFFF0000FFFF0000FFFF004"
    "B0F040006FFFF0F19000003E80164000003E801FF002800020000000000000000"
    "0000000000000000000000000000000F01CEFFFF03FFFFFF0000FFFF0000FFFF0"
    "000FFFF002D0F000006FFFF0F19000003E80164000003E8016400280002000000"
    "00000000000000000000000000000000000000000F00FAFFFF01CE10FF0000FFF"
    "F0000FFFF0000FFFF002D0F000006FFFF0F19000003E80164000003E801640028"
    "000500000000000000000000000000000000000000000000000F0000FFFF01FDF"
    "FFF03FFFFFF00FAFFFF01CE10FF0100E0A8"
)


@pytest.fixture
def mockD5NextDevice():
    device = _MockD5NextDevice()
    dev = Aquacomputer(
        device,
        "Mock Aquacomputer D5 Next",
        device_info=Aquacomputer._DEVICE_INFO[Aquacomputer._DEVICE_D5NEXT],
    )

    dev.connect()
    return dev


class _MockD5NextDevice(MockHidapiDevice):
    def __init__(self):
        super().__init__(vendor_id=0x0C70, product_id=0xF00E)

        self.preload_read(Report(1, D5NEXT_SAMPLE_STATUS_REPORT))
        self.preload_read(Report(3, D5NEXT_SAMPLE_CONTROL_REPORT))
        self.preload_read(Report(3, D5NEXT_SAMPLE_CONTROL_REPORT))

    def read(self, length):
        pre = super().read(length)
        if pre:
            return pre

        return Report(1, D5NEXT_SAMPLE_STATUS_REPORT)


def test_d5next_connect(mockD5NextDevice):
    def mock_open():
        nonlocal opened
        opened = True

    mockD5NextDevice.device.open = mock_open
    opened = False

    with mockD5NextDevice.connect() as cm:
        assert cm == mockD5NextDevice
        assert opened


def test_d5next_initialize(mockD5NextDevice):
    init_result = mockD5NextDevice.initialize()

    # Verify firmware version
    assert init_result[0][1] == 1023

    # Verify serial number
    assert init_result[1][1] == "03531-22908"


@pytest.mark.parametrize("has_hwmon,direct_access", [(False, False), (True, True)])
def test_d5next_get_status_directly(mockD5NextDevice, has_hwmon, direct_access):
    if has_hwmon:
        mockD5NextDevice._hwmon = HwmonDevice(None, None)

    got = mockD5NextDevice.get_status(direct_access=direct_access)

    expected = [
        ("Liquid temperature", pytest.approx(25.1, 0.1), "°C"),
        ("Soft. Sensor 1", pytest.approx(50, 0.1), "°C"),
        ("Pump speed", 1976, "rpm"),
        ("Pump power", pytest.approx(2.58, 0.1), "W"),
        ("Pump voltage", pytest.approx(12.02, 0.1), "V"),
        ("Pump current", pytest.approx(0.21, 0.1), "A"),
        ("Fan speed", 365, "rpm"),
        ("Fan power", pytest.approx(0.38, 0.1), "W"),
        ("Fan voltage", pytest.approx(12.04, 0.1), "V"),
        ("Fan current", pytest.approx(0.03, 0.1), "A"),
        ("+5V voltage", pytest.approx(5.00, 0.1), "V"),
        ("+12V voltage", pytest.approx(12.04, 0.1), "V"),
    ]

    assert sorted(got) == sorted(expected)


def test_d5next_get_status_from_hwmon(mockD5NextDevice, tmp_path):
    mockD5NextDevice._hwmon = HwmonDevice("mock_module", tmp_path)
    (tmp_path / "temp1_input").write_text("25100\n")  # Liquid temperature
    (tmp_path / "fan1_input").write_text("1976\n")  # Pump speed
    (tmp_path / "power1_input").write_text("2580000\n")  # Pump power
    (tmp_path / "in0_input").write_text("12020\n")  # Pump voltage
    (tmp_path / "curr1_input").write_text("215\n")  # Pump current
    (tmp_path / "fan2_input").write_text("365\n")  # Fan speed
    (tmp_path / "power2_input").write_text("380000\n")  # Fan power
    (tmp_path / "in1_input").write_text("12040\n")  # Fan voltage
    (tmp_path / "curr2_input").write_text("31\n")  # Fan current
    (tmp_path / "in2_input").write_text("4990\n")  # +5V voltage
    (tmp_path / "in3_input").write_text("12040\n")  # +12V voltage
    (tmp_path / "temp2_input").write_text("50000\n")  # Soft. Sensor 1 temperature
    (tmp_path / "temp3_input").write_text("60000\n")  # Soft. Sensor 2 temperature
    (tmp_path / "temp4_input").write_text("50000\n")  # Soft. Sensor 3 temperature
    (tmp_path / "temp5_input").write_text("50000\n")  # Soft. Sensor 4 temperature
    (tmp_path / "temp6_input").write_text("50000\n")  # Soft. Sensor 5 temperature
    (tmp_path / "temp7_input").write_text("50000\n")  # Soft. Sensor 6 temperature
    (tmp_path / "temp8_input").write_text("50000\n")  # Soft. Sensor 7 temperature
    (tmp_path / "temp9_input").write_text("50000\n")  # Soft. Sensor 8 temperature

    got = mockD5NextDevice.get_status()

    expected = [
        ("Liquid temperature", pytest.approx(25.1, 0.1), "°C"),
        ("Soft. Sensor 1", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 2", pytest.approx(60, 0.1), "°C"),
        ("Soft. Sensor 3", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 4", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 5", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 6", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 7", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 8", pytest.approx(50, 0.1), "°C"),
        ("Pump speed", 1976, "rpm"),
        ("Pump power", pytest.approx(2.58, 0.1), "W"),
        ("Pump voltage", pytest.approx(12.02, 0.1), "V"),
        ("Pump current", pytest.approx(0.21, 0.1), "A"),
        ("Fan speed", 365, "rpm"),
        ("Fan power", pytest.approx(0.38, 0.1), "W"),
        ("Fan voltage", pytest.approx(12.04, 0.1), "V"),
        ("Fan current", pytest.approx(0.03, 0.1), "A"),
        ("+5V voltage", pytest.approx(5.00, 0.1), "V"),
        ("+12V voltage", pytest.approx(12.04, 0.1), "V"),
    ]

    assert sorted(got) == sorted(expected)


@pytest.mark.parametrize("has_hwmon,direct_access", [(False, False), (True, True)])
def test_d5next_set_fixed_speeds_directly(mockD5NextDevice, has_hwmon, direct_access, tmp_path):
    """For both test cases only direct access should be used"""

    if has_hwmon:
        mockD5NextDevice._hwmon = HwmonDevice("mock_module", tmp_path)
        (tmp_path / "pwm1").write_text("0")
        (tmp_path / "pwm1_enable").write_text("0")
        (tmp_path / "pwm2").write_text("0")
        (tmp_path / "pwm2_enable").write_text("0")

    mockD5NextDevice.set_fixed_speed("pump", 84, direct_access=direct_access)
    mockD5NextDevice.set_fixed_speed("fan", 50, direct_access=direct_access)

    pump_report, fan_report = mockD5NextDevice.device.sent

    assert pump_report.number == 3
    assert pump_report.data[0x95:0x98] == [0, 32, 208]  # 0, <8400>
    assert fan_report.number == 3
    assert fan_report.data[0x40:0x43] == [0, 19, 136]  # 0, <5000>

    # Assert that hwmon wasn't touched
    if has_hwmon:
        assert (tmp_path / "pwm1_enable").read_text() == "0"
        assert (tmp_path / "pwm1").read_text() == "0"
        assert (tmp_path / "pwm2_enable").read_text() == "0"
        assert (tmp_path / "pwm2").read_text() == "0"


@pytest.mark.parametrize("has_support", [False, True])
def test_d5next_set_fixed_speeds_hwmon(mockD5NextDevice, has_support, tmp_path):
    mockD5NextDevice._hwmon = HwmonDevice("mock_module", tmp_path)

    if has_support:
        (tmp_path / "pwm1").write_text("0\n")
        (tmp_path / "pwm1_enable").write_text("0\n")
        (tmp_path / "pwm2").write_text("0\n")
        (tmp_path / "pwm2_enable").write_text("0\n")

    mockD5NextDevice.set_fixed_speed("pump", 84)
    mockD5NextDevice.set_fixed_speed("fan", 50)

    if has_support:
        assert (tmp_path / "pwm1_enable").read_text() == "1"
        assert (tmp_path / "pwm1").read_text() == "214"
        assert (tmp_path / "pwm2_enable").read_text() == "1"
        assert (tmp_path / "pwm2").read_text() == "127"
    else:
        # Assert fallback to direct access
        pump_report, fan_report = mockD5NextDevice.device.sent

        assert pump_report.number == 3
        assert pump_report.data[0x95:0x98] == [0, 32, 208]  # 0, <8400>
        assert fan_report.number == 3
        assert fan_report.data[0x40:0x43] == [0, 19, 136]  # 0, <5000>


def test_d5next_speed_profiles_not_supported(mockD5NextDevice):
    with pytest.raises(NotSupportedByDriver):
        mockD5NextDevice.set_speed_profile("fan", None)

    with pytest.raises(NotSupportedByDriver):
        mockD5NextDevice.set_speed_profile("pump", None)


@pytest.fixture
def mockFarbwerk360Device():
    device = _MockFarbwerk360Device()
    dev = Aquacomputer(
        device,
        "Mock Aquacomputer Farbwerk 360",
        device_info=Aquacomputer._DEVICE_INFO[Aquacomputer._DEVICE_FARBWERK360],
    )

    dev.connect()
    return dev


class _MockFarbwerk360Device(MockHidapiDevice):
    def __init__(self):
        super().__init__(vendor_id=0x0C70, product_id=0xF010)

        self.preload_read(Report(1, FARBWERK360_SAMPLE_STATUS_REPORT))

    def read(self, length):
        pre = super().read(length)
        if pre:
            return pre

        return Report(1, FARBWERK360_SAMPLE_STATUS_REPORT)


def test_farbwerk360_connect(mockFarbwerk360Device):
    def mock_open():
        nonlocal opened
        opened = True

    mockFarbwerk360Device.device.open = mock_open
    opened = False

    with mockFarbwerk360Device.connect() as cm:
        assert cm == mockFarbwerk360Device
        assert opened


def test_farbwerk360_initialize(mockFarbwerk360Device):
    init_result = mockFarbwerk360Device.initialize()

    # Verify firmware version
    assert init_result[0][1] == 1022

    # Verify serial number
    assert init_result[1][1] == "16827-56978"


@pytest.mark.parametrize("has_hwmon,direct_access", [(False, False), (True, True)])
def test_farbwerk360_get_status_directly(mockFarbwerk360Device, has_hwmon, direct_access):
    if has_hwmon:
        mockFarbwerk360Device._hwmon = HwmonDevice(None, None)

    got = mockFarbwerk360Device.get_status(direct_access=direct_access)

    expected = [
        ("Sensor 1", pytest.approx(26.2, 0.1), "°C"),
        ("Sensor 2", pytest.approx(26.3, 0.1), "°C"),
        ("Sensor 3", pytest.approx(26.7, 0.1), "°C"),
        ("Sensor 4", pytest.approx(25.5, 0.1), "°C"),
        ("Soft. Sensor 1", pytest.approx(52.00, 0.1), "°C"),
    ]

    assert sorted(got) == sorted(expected)


def test_farbwerk360_get_status_from_hwmon(mockFarbwerk360Device, tmp_path):
    mockFarbwerk360Device._hwmon = HwmonDevice("mock_module", tmp_path)
    (tmp_path / "temp1_input").write_text("26200\n")
    (tmp_path / "temp2_input").write_text("26310\n")
    (tmp_path / "temp3_input").write_text("26710\n")
    (tmp_path / "temp4_input").write_text("25520\n")
    (tmp_path / "temp5_input").write_text("50000\n")  # Soft. Sensor 1 temperature
    (tmp_path / "temp6_input").write_text("60000\n")  # Soft. Sensor 2 temperature
    (tmp_path / "temp7_input").write_text("50000\n")  # Soft. Sensor 3 temperature
    (tmp_path / "temp8_input").write_text("50000\n")  # Soft. Sensor 4 temperature
    (tmp_path / "temp9_input").write_text("50000\n")  # Soft. Sensor 5 temperature
    (tmp_path / "temp10_input").write_text("50000\n")  # Soft. Sensor 6 temperature
    (tmp_path / "temp11_input").write_text("50000\n")  # Soft. Sensor 7 temperature
    (tmp_path / "temp12_input").write_text("50000\n")  # Soft. Sensor 8 temperature
    (tmp_path / "temp13_input").write_text("50000\n")  # Soft. Sensor 9 temperature
    (tmp_path / "temp14_input").write_text("50000\n")  # Soft. Sensor 10 temperature
    (tmp_path / "temp15_input").write_text("50000\n")  # Soft. Sensor 11 temperature
    (tmp_path / "temp16_input").write_text("50000\n")  # Soft. Sensor 12 temperature
    (tmp_path / "temp17_input").write_text("50000\n")  # Soft. Sensor 13 temperature
    (tmp_path / "temp18_input").write_text("50000\n")  # Soft. Sensor 14 temperature
    (tmp_path / "temp19_input").write_text("50000\n")  # Soft. Sensor 15 temperature
    (tmp_path / "temp20_input").write_text("50000\n")  # Soft. Sensor 16 temperature

    got = mockFarbwerk360Device.get_status()

    expected = [
        ("Sensor 1", pytest.approx(26.2, 0.1), "°C"),
        ("Sensor 2", pytest.approx(26.3, 0.1), "°C"),
        ("Sensor 3", pytest.approx(26.7, 0.1), "°C"),
        ("Sensor 4", pytest.approx(25.5, 0.1), "°C"),
        ("Soft. Sensor 1", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 2", pytest.approx(60, 0.1), "°C"),
        ("Soft. Sensor 3", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 4", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 5", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 6", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 7", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 8", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 9", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 10", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 11", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 12", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 13", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 14", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 15", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 16", pytest.approx(50, 0.1), "°C"),
    ]

    assert sorted(got) == sorted(expected)


def test_farbwerk360_set_fixed_speeds_not_supported(mockFarbwerk360Device):
    with pytest.raises(NotSupportedByDevice):
        mockFarbwerk360Device.set_fixed_speed("fan", 42)


def test_farbwerk360_speed_profiles_not_supported(mockFarbwerk360Device):
    with pytest.raises(NotSupportedByDevice):
        mockFarbwerk360Device.set_speed_profile("fan", None)


@pytest.fixture
def mockOctoDevice():
    device = _MockOctoDevice()
    dev = Aquacomputer(
        device,
        "Mock Aquacomputer Octo",
        device_info=Aquacomputer._DEVICE_INFO[Aquacomputer._DEVICE_OCTO],
    )

    dev.connect()
    return dev


class _MockOctoDevice(MockHidapiDevice):
    def __init__(self):
        super().__init__(vendor_id=0x0C70, product_id=0xF011)

        self.preload_read(Report(1, OCTO_SAMPLE_STATUS_REPORT))
        self.preload_read(Report(3, OCTO_SAMPLE_CONTROL_REPORT))
        self.preload_read(Report(3, OCTO_SAMPLE_CONTROL_REPORT))

    def read(self, length):
        pre = super().read(length)
        if pre:
            return pre

        return Report(1, OCTO_SAMPLE_STATUS_REPORT)


def test_octo_connect(mockOctoDevice):
    def mock_open():
        nonlocal opened
        opened = True

    mockOctoDevice.device.open = mock_open
    opened = False

    with mockOctoDevice.connect() as cm:
        assert cm == mockOctoDevice
        assert opened


def test_octo_initialize(mockOctoDevice):
    init_result = mockOctoDevice.initialize()

    # Verify firmware version
    assert init_result[0][1] == 1019

    # Verify serial number
    assert init_result[1][1] == "14994-51690"


@pytest.mark.parametrize("has_hwmon,direct_access", [(False, False), (True, True)])
def test_octo_get_status_directly(mockOctoDevice, has_hwmon, direct_access):
    if has_hwmon:
        mockOctoDevice._hwmon = HwmonDevice(None, None)

    got = mockOctoDevice.get_status(direct_access=direct_access)

    expected = [
        ("Sensor 1", pytest.approx(27.5, 0.1), "°C"),
        ("Sensor 2", pytest.approx(27.7, 0.1), "°C"),
        ("Sensor 3", pytest.approx(28.4, 0.1), "°C"),
        ("Sensor 4", pytest.approx(34.2, 0.1), "°C"),
        ("Soft. Sensor 1", pytest.approx(37.84, 0.1), "°C"),
        ("Fan 1 speed", pytest.approx(0, 0.1), "rpm"),
        ("Fan 1 power", pytest.approx(0.01, 0.1), "W"),
        ("Fan 1 voltage", pytest.approx(12.09, 0.1), "V"),
        ("Fan 1 current", pytest.approx(0.001, 0.1), "A"),
        ("Fan 2 speed", pytest.approx(576, 0.1), "rpm"),
        ("Fan 2 power", pytest.approx(1.03, 0.1), "W"),
        ("Fan 2 voltage", pytest.approx(12.09, 0.1), "V"),
        ("Fan 2 current", pytest.approx(0.35, 0.1), "A"),
        ("Fan 3 speed", pytest.approx(0, 0.1), "rpm"),
        ("Fan 3 power", pytest.approx(0, 0.1), "W"),
        ("Fan 3 voltage", pytest.approx(0, 0.1), "V"),
        ("Fan 3 current", pytest.approx(0, 0.1), "A"),
        ("Fan 4 speed", pytest.approx(0, 0.1), "rpm"),
        ("Fan 4 power", pytest.approx(0, 0.1), "W"),
        ("Fan 4 voltage", pytest.approx(0, 0.1), "V"),
        ("Fan 4 current", pytest.approx(0, 0.1), "A"),
        ("Fan 5 speed", pytest.approx(0, 0.1), "rpm"),
        ("Fan 5 power", pytest.approx(0, 0.1), "W"),
        ("Fan 5 voltage", pytest.approx(0, 0.1), "V"),
        ("Fan 5 current", pytest.approx(0, 0.1), "A"),
        ("Fan 6 speed", pytest.approx(0, 0.1), "rpm"),
        ("Fan 6 power", pytest.approx(0, 0.1), "W"),
        ("Fan 6 voltage", pytest.approx(0, 0.1), "V"),
        ("Fan 6 current", pytest.approx(0, 0.1), "A"),
        ("Fan 7 speed", pytest.approx(0, 0.1), "rpm"),
        ("Fan 7 power", pytest.approx(0, 0.1), "W"),
        ("Fan 7 voltage", pytest.approx(0, 0.1), "V"),
        ("Fan 7 current", pytest.approx(0, 0.1), "A"),
        ("Fan 8 speed", pytest.approx(0, 0.1), "rpm"),
        ("Fan 8 power", pytest.approx(0.02, 0.1), "W"),
        ("Fan 8 voltage", pytest.approx(12.09, 0.1), "V"),
        ("Fan 8 current", pytest.approx(0.002, 0.1), "A"),
    ]

    assert sorted(got) == sorted(expected)


def test_octo_get_status_from_hwmon(mockOctoDevice, tmp_path):
    mockOctoDevice._hwmon = HwmonDevice("mock_module", tmp_path)
    (tmp_path / "temp1_input").write_text("27580\n")
    (tmp_path / "temp2_input").write_text("27670\n")
    (tmp_path / "temp3_input").write_text("28370\n")
    (tmp_path / "temp4_input").write_text("34240\n")
    (tmp_path / "fan1_input").write_text("0\n")
    (tmp_path / "power1_input").write_text("10000\n")
    (tmp_path / "in0_input").write_text("12090\n")
    (tmp_path / "curr1_input").write_text("1\n")
    (tmp_path / "fan2_input").write_text("576\n")
    (tmp_path / "power2_input").write_text("1030000\n")
    (tmp_path / "in1_input").write_text("12090\n")
    (tmp_path / "curr2_input").write_text("350\n")
    (tmp_path / "fan3_input").write_text("0\n")
    (tmp_path / "power3_input").write_text("0\n")
    (tmp_path / "in2_input").write_text("0\n")
    (tmp_path / "curr3_input").write_text("0\n")
    (tmp_path / "fan4_input").write_text("0\n")
    (tmp_path / "power4_input").write_text("0\n")
    (tmp_path / "in3_input").write_text("0\n")
    (tmp_path / "curr4_input").write_text("0\n")
    (tmp_path / "fan5_input").write_text("0\n")
    (tmp_path / "power5_input").write_text("0\n")
    (tmp_path / "in4_input").write_text("0\n")
    (tmp_path / "curr5_input").write_text("0\n")
    (tmp_path / "fan6_input").write_text("0\n")
    (tmp_path / "power6_input").write_text("0\n")
    (tmp_path / "in5_input").write_text("0\n")
    (tmp_path / "curr6_input").write_text("0\n")
    (tmp_path / "fan7_input").write_text("0\n")
    (tmp_path / "power7_input").write_text("0\n")
    (tmp_path / "in6_input").write_text("0\n")
    (tmp_path / "curr7_input").write_text("0\n")
    (tmp_path / "fan8_input").write_text("0\n")
    (tmp_path / "power8_input").write_text("20000\n")
    (tmp_path / "in7_input").write_text("12090\n")
    (tmp_path / "curr8_input").write_text("2\n")
    (tmp_path / "temp5_input").write_text("50000\n")  # Soft. Sensor 1 temperature
    (tmp_path / "temp6_input").write_text("60000\n")  # Soft. Sensor 2 temperature
    (tmp_path / "temp7_input").write_text("50000\n")  # Soft. Sensor 3 temperature
    (tmp_path / "temp8_input").write_text("50000\n")  # Soft. Sensor 4 temperature
    (tmp_path / "temp9_input").write_text("50000\n")  # Soft. Sensor 5 temperature
    (tmp_path / "temp10_input").write_text("50000\n")  # Soft. Sensor 6 temperature
    (tmp_path / "temp11_input").write_text("50000\n")  # Soft. Sensor 7 temperature
    (tmp_path / "temp12_input").write_text("50000\n")  # Soft. Sensor 8 temperature
    (tmp_path / "temp13_input").write_text("50000\n")  # Soft. Sensor 9 temperature
    (tmp_path / "temp14_input").write_text("50000\n")  # Soft. Sensor 10 temperature
    (tmp_path / "temp15_input").write_text("50000\n")  # Soft. Sensor 11 temperature
    (tmp_path / "temp16_input").write_text("50000\n")  # Soft. Sensor 12 temperature
    (tmp_path / "temp17_input").write_text("50000\n")  # Soft. Sensor 13 temperature
    (tmp_path / "temp18_input").write_text("50000\n")  # Soft. Sensor 14 temperature
    (tmp_path / "temp19_input").write_text("50000\n")  # Soft. Sensor 15 temperature
    (tmp_path / "temp20_input").write_text("50000\n")  # Soft. Sensor 16 temperature

    got = mockOctoDevice.get_status()

    expected = [
        ("Sensor 1", pytest.approx(27.5, 0.1), "°C"),
        ("Sensor 2", pytest.approx(27.7, 0.1), "°C"),
        ("Sensor 3", pytest.approx(28.4, 0.1), "°C"),
        ("Sensor 4", pytest.approx(34.2, 0.1), "°C"),
        ("Fan 1 speed", pytest.approx(0, 0.1), "rpm"),
        ("Fan 1 power", pytest.approx(0.01, 0.1), "W"),
        ("Fan 1 voltage", pytest.approx(12.09, 0.1), "V"),
        ("Fan 1 current", pytest.approx(0.001, 0.1), "A"),
        ("Fan 2 speed", pytest.approx(576, 0.1), "rpm"),
        ("Fan 2 power", pytest.approx(1.03, 0.1), "W"),
        ("Fan 2 voltage", pytest.approx(12.09, 0.1), "V"),
        ("Fan 2 current", pytest.approx(0.35, 0.1), "A"),
        ("Fan 3 speed", pytest.approx(0, 0.1), "rpm"),
        ("Fan 3 power", pytest.approx(0, 0.1), "W"),
        ("Fan 3 voltage", pytest.approx(0, 0.1), "V"),
        ("Fan 3 current", pytest.approx(0, 0.1), "A"),
        ("Fan 4 speed", pytest.approx(0, 0.1), "rpm"),
        ("Fan 4 power", pytest.approx(0, 0.1), "W"),
        ("Fan 4 voltage", pytest.approx(0, 0.1), "V"),
        ("Fan 4 current", pytest.approx(0, 0.1), "A"),
        ("Fan 5 speed", pytest.approx(0, 0.1), "rpm"),
        ("Fan 5 power", pytest.approx(0, 0.1), "W"),
        ("Fan 5 voltage", pytest.approx(0, 0.1), "V"),
        ("Fan 5 current", pytest.approx(0, 0.1), "A"),
        ("Fan 6 speed", pytest.approx(0, 0.1), "rpm"),
        ("Fan 6 power", pytest.approx(0, 0.1), "W"),
        ("Fan 6 voltage", pytest.approx(0, 0.1), "V"),
        ("Fan 6 current", pytest.approx(0, 0.1), "A"),
        ("Fan 7 speed", pytest.approx(0, 0.1), "rpm"),
        ("Fan 7 power", pytest.approx(0, 0.1), "W"),
        ("Fan 7 voltage", pytest.approx(0, 0.1), "V"),
        ("Fan 7 current", pytest.approx(0, 0.1), "A"),
        ("Fan 8 speed", pytest.approx(0, 0.1), "rpm"),
        ("Fan 8 power", pytest.approx(0.02, 0.1), "W"),
        ("Fan 8 voltage", pytest.approx(12.09, 0.1), "V"),
        ("Fan 8 current", pytest.approx(0.002, 0.1), "A"),
        ("Soft. Sensor 1", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 2", pytest.approx(60, 0.1), "°C"),
        ("Soft. Sensor 3", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 4", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 5", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 6", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 7", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 8", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 9", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 10", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 11", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 12", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 13", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 14", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 15", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 16", pytest.approx(50, 0.1), "°C"),
    ]

    assert sorted(got) == sorted(expected)


@pytest.mark.parametrize("has_hwmon,direct_access", [(False, False), (True, True)])
def test_octo_set_fixed_speeds_directly(mockOctoDevice, has_hwmon, direct_access, tmp_path):
    """For both test cases only direct access should be used"""

    if has_hwmon:
        mockOctoDevice._hwmon = HwmonDevice("mock_module", tmp_path)
        (tmp_path / "pwm1").write_text("0")
        (tmp_path / "pwm1_enable").write_text("0")
        (tmp_path / "pwm2").write_text("0")
        (tmp_path / "pwm2_enable").write_text("0")

    mockOctoDevice.set_fixed_speed("fan1", 84, direct_access=direct_access)
    mockOctoDevice.set_fixed_speed("fan2", 50, direct_access=direct_access)

    pump_report, fan_report = mockOctoDevice.device.sent

    assert pump_report.number == 3
    assert pump_report.data[0x59:0x5C] == [0, 32, 208]  # 0, <8400>
    assert fan_report.number == 3
    assert fan_report.data[0xAE:0xB1] == [0, 19, 136]  # 0, <5000>

    # Assert that hwmon wasn't touched
    if has_hwmon:
        assert (tmp_path / "pwm1_enable").read_text() == "0"
        assert (tmp_path / "pwm1").read_text() == "0"
        assert (tmp_path / "pwm2_enable").read_text() == "0"
        assert (tmp_path / "pwm2").read_text() == "0"


@pytest.mark.parametrize("has_support", [False, True])
def test_octo_set_fixed_speeds_hwmon(mockOctoDevice, has_support, tmp_path):
    mockOctoDevice._hwmon = HwmonDevice("mock_module", tmp_path)

    if has_support:
        (tmp_path / "pwm1").write_text("0\n")
        (tmp_path / "pwm1_enable").write_text("0\n")
        (tmp_path / "pwm2").write_text("0\n")
        (tmp_path / "pwm2_enable").write_text("0\n")

    mockOctoDevice.set_fixed_speed("fan1", 84)
    mockOctoDevice.set_fixed_speed("fan2", 50)

    if has_support:
        assert (tmp_path / "pwm1_enable").read_text() == "1"
        assert (tmp_path / "pwm1").read_text() == "214"
        assert (tmp_path / "pwm2_enable").read_text() == "1"
        assert (tmp_path / "pwm2").read_text() == "127"
    else:
        # Assert fallback to direct access
        pump_report, fan_report = mockOctoDevice.device.sent

        assert pump_report.number == 3
        assert pump_report.data[0x59:0x5C] == [0, 32, 208]  # 0, <8400>
        assert fan_report.number == 3
        assert fan_report.data[0xAE:0xB1] == [0, 19, 136]  # 0, <5000>


def test_octo_speed_profiles_not_supported(mockOctoDevice):
    with pytest.raises(NotSupportedByDriver):
        mockOctoDevice.set_speed_profile("fan", None)


@pytest.fixture
def mockQuadroDevice():
    device = _MockQuadroDevice()
    dev = Aquacomputer(
        device,
        "Mock Aquacomputer Quadro",
        device_info=Aquacomputer._DEVICE_INFO[Aquacomputer._DEVICE_QUADRO],
    )

    dev.connect()
    return dev


class _MockQuadroDevice(MockHidapiDevice):
    def __init__(self):
        super().__init__(vendor_id=0x0C70, product_id=0xF00D)

        self.preload_read(Report(1, QUADRO_SAMPLE_STATUS_REPORT))
        self.preload_read(Report(3, QUADRO_SAMPLE_CONTROL_REPORT))
        self.preload_read(Report(3, QUADRO_SAMPLE_CONTROL_REPORT))

    def read(self, length):
        pre = super().read(length)
        if pre:
            return pre

        return Report(1, QUADRO_SAMPLE_STATUS_REPORT)


def test_quadro_connect(mockQuadroDevice):
    def mock_open():
        nonlocal opened
        opened = True

    mockQuadroDevice.device.open = mock_open
    opened = False

    with mockQuadroDevice.connect() as cm:
        assert cm == mockQuadroDevice
        assert opened


def test_quadro_initialize(mockQuadroDevice):
    init_result = mockQuadroDevice.initialize()

    # Verify firmware version
    assert init_result[0][1] == 1032

    # Verify serial number
    assert init_result[1][1] == "23410-65344"


@pytest.mark.parametrize("has_hwmon,direct_access", [(False, False), (True, True)])
def test_quadro_get_status_directly(mockQuadroDevice, has_hwmon, direct_access):
    if has_hwmon:
        mockQuadroDevice._hwmon = HwmonDevice(None, None)

    got = mockQuadroDevice.get_status(direct_access=direct_access)

    expected = [
        ("Sensor 3", pytest.approx(16.17, 0.1), "°C"),
        ("Soft. Sensor 1", pytest.approx(23.9, 0.1), "°C"),
        ("Soft. Sensor 13", pytest.approx(50.0, 0.1), "°C"),
        ("Fan 1 speed", pytest.approx(0, 0.1), "rpm"),
        ("Fan 1 power", pytest.approx(0, 0.1), "W"),
        ("Fan 1 voltage", pytest.approx(0, 0.1), "V"),
        ("Fan 1 current", pytest.approx(0, 0.1), "A"),
        ("Fan 2 speed", pytest.approx(0, 0.1), "rpm"),
        ("Fan 2 power", pytest.approx(0, 0.1), "W"),
        ("Fan 2 voltage", pytest.approx(12.07, 0.1), "V"),
        ("Fan 2 current", pytest.approx(0, 0.1), "A"),
        ("Fan 3 speed", pytest.approx(356, 0.1), "rpm"),
        ("Fan 3 power", pytest.approx(0, 0.1), "W"),
        ("Fan 3 voltage", pytest.approx(12.07, 0.1), "V"),
        ("Fan 3 current", pytest.approx(0, 0.1), "A"),
        ("Fan 4 speed", pytest.approx(0, 0.1), "rpm"),
        ("Fan 4 power", pytest.approx(0, 0.1), "W"),
        ("Fan 4 voltage", pytest.approx(12.07, 0.1), "V"),
        ("Fan 4 current", pytest.approx(0, 0.1), "A"),
        ("Flow sensor", pytest.approx(0, 0.1), "dL/h"),
    ]

    assert sorted(got) == sorted(expected)


def test_quadro_get_status_from_hwmon(mockQuadroDevice, tmp_path):
    mockQuadroDevice._hwmon = HwmonDevice("mock_module", tmp_path)
    (tmp_path / "temp1_input").write_text("27580\n")
    (tmp_path / "temp2_input").write_text("27670\n")
    (tmp_path / "temp3_input").write_text("28370\n")
    (tmp_path / "temp4_input").write_text("34240\n")
    (tmp_path / "fan1_input").write_text("0\n")
    (tmp_path / "power1_input").write_text("10000\n")
    (tmp_path / "in0_input").write_text("12090\n")
    (tmp_path / "curr1_input").write_text("1\n")
    (tmp_path / "fan2_input").write_text("576\n")
    (tmp_path / "power2_input").write_text("1030000\n")
    (tmp_path / "in1_input").write_text("12090\n")
    (tmp_path / "curr2_input").write_text("350\n")
    (tmp_path / "fan3_input").write_text("0\n")
    (tmp_path / "power3_input").write_text("0\n")
    (tmp_path / "in2_input").write_text("0\n")
    (tmp_path / "curr3_input").write_text("0\n")
    (tmp_path / "fan4_input").write_text("0\n")
    (tmp_path / "power4_input").write_text("0\n")
    (tmp_path / "in3_input").write_text("0\n")
    (tmp_path / "curr4_input").write_text("0\n")
    (tmp_path / "fan5_input").write_text("603\n")
    (tmp_path / "temp5_input").write_text("50000\n")  # Soft. Sensor 1 temperature
    (tmp_path / "temp6_input").write_text("60000\n")  # Soft. Sensor 2 temperature
    (tmp_path / "temp7_input").write_text("50000\n")  # Soft. Sensor 3 temperature
    (tmp_path / "temp8_input").write_text("50000\n")  # Soft. Sensor 4 temperature
    (tmp_path / "temp9_input").write_text("50000\n")  # Soft. Sensor 5 temperature
    (tmp_path / "temp10_input").write_text("50000\n")  # Soft. Sensor 6 temperature
    (tmp_path / "temp11_input").write_text("50000\n")  # Soft. Sensor 7 temperature
    (tmp_path / "temp12_input").write_text("50000\n")  # Soft. Sensor 8 temperature
    (tmp_path / "temp13_input").write_text("50000\n")  # Soft. Sensor 9 temperature
    (tmp_path / "temp14_input").write_text("50000\n")  # Soft. Sensor 10 temperature
    (tmp_path / "temp15_input").write_text("50000\n")  # Soft. Sensor 11 temperature
    (tmp_path / "temp16_input").write_text("50000\n")  # Soft. Sensor 12 temperature
    (tmp_path / "temp17_input").write_text("50000\n")  # Soft. Sensor 13 temperature
    (tmp_path / "temp18_input").write_text("50000\n")  # Soft. Sensor 14 temperature
    (tmp_path / "temp19_input").write_text("50000\n")  # Soft. Sensor 15 temperature
    (tmp_path / "temp20_input").write_text("50000\n")  # Soft. Sensor 16 temperature

    got = mockQuadroDevice.get_status()

    expected = [
        ("Sensor 1", pytest.approx(27.5, 0.1), "°C"),
        ("Sensor 2", pytest.approx(27.7, 0.1), "°C"),
        ("Sensor 3", pytest.approx(28.4, 0.1), "°C"),
        ("Sensor 4", pytest.approx(34.2, 0.1), "°C"),
        ("Fan 1 speed", pytest.approx(0, 0.1), "rpm"),
        ("Fan 1 power", pytest.approx(0.01, 0.1), "W"),
        ("Fan 1 voltage", pytest.approx(12.09, 0.1), "V"),
        ("Fan 1 current", pytest.approx(0.001, 0.1), "A"),
        ("Fan 2 speed", pytest.approx(576, 0.1), "rpm"),
        ("Fan 2 power", pytest.approx(1.03, 0.1), "W"),
        ("Fan 2 voltage", pytest.approx(12.09, 0.1), "V"),
        ("Fan 2 current", pytest.approx(0.35, 0.1), "A"),
        ("Fan 3 speed", pytest.approx(0, 0.1), "rpm"),
        ("Fan 3 power", pytest.approx(0, 0.1), "W"),
        ("Fan 3 voltage", pytest.approx(0, 0.1), "V"),
        ("Fan 3 current", pytest.approx(0, 0.1), "A"),
        ("Fan 4 speed", pytest.approx(0, 0.1), "rpm"),
        ("Fan 4 power", pytest.approx(0, 0.1), "W"),
        ("Fan 4 voltage", pytest.approx(0, 0.1), "V"),
        ("Fan 4 current", pytest.approx(0, 0.1), "A"),
        ("Flow sensor", pytest.approx(603, 0.1), "dL/h"),
        ("Soft. Sensor 1", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 2", pytest.approx(60, 0.1), "°C"),
        ("Soft. Sensor 3", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 4", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 5", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 6", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 7", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 8", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 9", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 10", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 11", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 12", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 13", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 14", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 15", pytest.approx(50, 0.1), "°C"),
        ("Soft. Sensor 16", pytest.approx(50, 0.1), "°C"),
    ]

    assert sorted(got) == sorted(expected)


@pytest.mark.parametrize("has_hwmon,direct_access", [(False, False), (True, True)])
def test_quadro_set_fixed_speeds_directly(mockQuadroDevice, has_hwmon, direct_access, tmp_path):
    """For both test cases only direct access should be used"""

    if has_hwmon:
        mockQuadroDevice._hwmon = HwmonDevice("mock_module", tmp_path)
        (tmp_path / "pwm1").write_text("0")
        (tmp_path / "pwm1_enable").write_text("0")
        (tmp_path / "pwm2").write_text("0")
        (tmp_path / "pwm2_enable").write_text("0")

    mockQuadroDevice.set_fixed_speed("fan1", 84, direct_access=direct_access)
    mockQuadroDevice.set_fixed_speed("fan2", 50, direct_access=direct_access)

    pump_report, fan_report = mockQuadroDevice.device.sent

    assert pump_report.number == 3
    assert pump_report.data[0x35:0x38] == [0, 32, 208]  # 0, <8400>
    assert fan_report.number == 3
    assert fan_report.data[0x8A:0x8D] == [0, 19, 136]  # 0, <5000>

    # Assert that hwmon wasn't touched
    if has_hwmon:
        assert (tmp_path / "pwm1_enable").read_text() == "0"
        assert (tmp_path / "pwm1").read_text() == "0"
        assert (tmp_path / "pwm2_enable").read_text() == "0"
        assert (tmp_path / "pwm2").read_text() == "0"


@pytest.mark.parametrize("has_support", [False, True])
def test_quadro_set_fixed_speeds_hwmon(mockQuadroDevice, has_support, tmp_path):
    mockQuadroDevice._hwmon = HwmonDevice("mock_module", tmp_path)

    if has_support:
        (tmp_path / "pwm1").write_text("0\n")
        (tmp_path / "pwm1_enable").write_text("0\n")
        (tmp_path / "pwm2").write_text("0\n")
        (tmp_path / "pwm2_enable").write_text("0\n")

    mockQuadroDevice.set_fixed_speed("fan1", 84)
    mockQuadroDevice.set_fixed_speed("fan2", 50)

    if has_support:
        assert (tmp_path / "pwm1_enable").read_text() == "1"
        assert (tmp_path / "pwm1").read_text() == "214"
        assert (tmp_path / "pwm2_enable").read_text() == "1"
        assert (tmp_path / "pwm2").read_text() == "127"
    else:
        # Assert fallback to direct access
        pump_report, fan_report = mockQuadroDevice.device.sent

        assert pump_report.number == 3
        assert pump_report.data[0x35:0x38] == [0, 32, 208]  # 0, <8400>
        assert fan_report.number == 3
        assert fan_report.data[0x8A:0x8D] == [0, 19, 136]  # 0, <5000>


def test_quadro_speed_profiles_not_supported(mockQuadroDevice):
    with pytest.raises(NotSupportedByDriver):
        mockQuadroDevice.set_speed_profile("fan", None)



"""liquidctl drivers for Corsair Commander Core.

Supported devices:

- Corsair Commander Core

Copyright ParkerMc and contributors
SPDX-License-Identifier: GPL-3.0-or-later
"""

import logging
from contextlib import contextmanager

from liquidctl.driver.usb import UsbHidDriver
from liquidctl.error import ExpectationNotMet, NotSupportedByDriver, NotSupportedByDevice
from liquidctl.util import clamp, u16le_from

_LOGGER = logging.getLogger(__name__)

_REPORT_LENGTH = 96
_RESPONSE_LENGTH = 96

_INTERFACE_NUMBER = 0

_FAN_COUNT = 6

_CMD_WAKE = (0x01, 0x03, 0x00, 0x02)
_CMD_SLEEP = (0x01, 0x03, 0x00, 0x01)
_CMD_GET_FIRMWARE = (0x02, 0x13)
_CMD_CLOSE_ENDPOINT = (0x05, 0x01, 0x00)
_CMD_OPEN_ENDPOINT = (0x0d, 0x00)
_CMD_READ = (0x08, 0x00)
_CMD_WRITE = (0x06, 0x00)

_MODE_LED_COUNT = (0x20,)
_MODE_GET_SPEEDS = (0x17,)
_MODE_GET_TEMPS = (0x21,)
_MODE_CONNECTED_SPEEDS = (0x1a,)
_MODE_HW_SPEED_MODE = (0x60, 0x6d)
_MODE_HW_FIXED_PERCENT = (0x61, 0x6d)

_DATA_TYPE_SPEEDS = (0x06, 0x00)
_DATA_TYPE_LED_COUNT = (0x0f, 0x00)
_DATA_TYPE_TEMPS = (0x10, 0x00)
_DATA_TYPE_CONNECTED_SPEEDS = (0x09, 0x00)
_DATA_TYPE_HW_SPEED_MODE = (0x03, 0x00)
_DATA_TYPE_HW_FIXED_PERCENT = (0x04, 0x00)


class CommanderCore(UsbHidDriver):
    """Corsair Commander Core"""

    _MATCHES = [
        (0x1b1c, 0x0c1c, 'Corsair Commander Core (experimental)', {"has_pump": True}),
        (0x1b1c, 0x0c2a, 'Corsair Commander Core XT (experimental)', {"has_pump": False}),
        (0x1b1c, 0x0c32, 'Corsair Commander ST (experimental)', {"has_pump": True}),
    ]

    def __init__(self, device, description, has_pump, **kwargs):
        super().__init__(device, description, **kwargs)
        self._has_pump = has_pump

    def initialize(self, **kwargs):
        """Initialize the device and get the fan modes."""

        with self._wake_device_context():
            # Get Firmware
            res = self._send_command(_CMD_GET_FIRMWARE)
            fw_version = (res[3], res[4], res[5])

            status = [('Firmware version', '{}.{}.{}'.format(*fw_version), '')]

            # Get LEDs per fan
            res = self._read_data(_MODE_LED_COUNT, _DATA_TYPE_LED_COUNT)
            num_devices = res[0]
            led_data = res[1:1 + num_devices * 4]
            for i in range(0, num_devices):
                connected = u16le_from(led_data, offset=i * 4) == 2
                num_leds = u16le_from(led_data, offset=i * 4 + 2)
                if self._has_pump:
                    label = 'AIO LED count' if i == 0 else f'RGB port {i} LED count'
                else:
                    label = f'RGB port {i+1} LED count'

                status += [(label, num_leds if connected else None, '')]

            # Get what fans are connected
            res = self._read_data(_MODE_CONNECTED_SPEEDS, _DATA_TYPE_CONNECTED_SPEEDS)
            num_devices = res[0]
            for i in range(0, num_devices):
                if self._has_pump:
                    label = 'AIO port connected' if i == 0 else f'Fan port {i} connected'
                else:
                    label = f'Fan port {i+1} connected'

                status += [(label, res[i + 1] == 0x07, '')]

            # Get what temp sensors are connected
            for i, temp in enumerate(self._get_temps()):
                connected = temp is not None
                if self._has_pump:
                    label = 'Water temperature sensor' if i == 0 and self._has_pump else f'Temperature sensor {i}'
                else:
                    label = f'Temperature sensor {i+1}'

                status += [(label, connected, '')]

        return status

    def get_status(self, **kwargs):
        """Get all the fan speeds and temps"""
        status = []

        with self._wake_device_context():
            for i, speed in enumerate(self._get_speeds()):
                if self._has_pump:
                    label = 'Pump speed' if i == 0 else f'Fan speed {i}'
                else:
                    label = f'Fan speed {i+1}'

                status += [(label, speed, 'rpm')]

            for i, temp in enumerate(self._get_temps()):
                if temp is None:
                    continue

                if self._has_pump:
                    label = 'Water temperature' if i == 0 else f'Temperature {i}'
                else:
                    label = f'Temperature {i}'

                status += [(label, temp, '°C')]

        return status

    def set_color(self, channel, mode, colors, **kwargs):
        raise NotSupportedByDriver

    def set_speed_profile(self, channel, profile, **kwargs):
        raise NotSupportedByDriver

    def set_fixed_speed(self, channel, duty, **kwargs):
        channels = self._parse_channels(channel)

        with self._wake_device_context():
            # Set hardware speed mode
            res = self._read_data(_MODE_HW_SPEED_MODE, _DATA_TYPE_HW_SPEED_MODE)
            device_count = res[0]

            data = bytearray(res[0:device_count + 1])
            for chan in channels:
                data[chan + 1] = 0x00  # Set the device's hardware mode to fixed percent
            self._write_data(_MODE_HW_SPEED_MODE, _DATA_TYPE_HW_SPEED_MODE, data)

            # Set speed
            res = self._read_data(_MODE_HW_FIXED_PERCENT, _DATA_TYPE_HW_FIXED_PERCENT)
            device_count = res[0]
            data = bytearray(res[0:device_count * 2 + 1])
            duty_le = int.to_bytes(clamp(duty, 0, 100), length=2, byteorder="little", signed=False)
            for chan in channels:
                i = chan * 2 + 1
                data[i: i + 2] = duty_le  # Update the device speed
            self._write_data(_MODE_HW_FIXED_PERCENT, _DATA_TYPE_HW_FIXED_PERCENT, data)

    @classmethod
    def probe(cls, handle, **kwargs):
        """Ensure we get the right interface"""

        if handle.hidinfo['interface_number'] != _INTERFACE_NUMBER:
            return

        yield from super().probe(handle, **kwargs)

    def _get_speeds(self):
        speeds = []

        res = self._read_data(_MODE_GET_SPEEDS, _DATA_TYPE_SPEEDS)

        num_speeds = res[0]
        speeds_data = res[1:1 + num_speeds * 2]
        for i in range(0, num_speeds):
            speeds.append(u16le_from(speeds_data, offset=i * 2))

        return speeds

    def _get_temps(self):
        temps = []

        res = self._read_data(_MODE_GET_TEMPS, _DATA_TYPE_TEMPS)

        num_temps = res[0]
        temp_data = res[1:1 + num_temps * 3]
        for i in range(0, num_temps):
            connected = temp_data[i * 3] == 0x00
            if connected:
                temps.append(u16le_from(temp_data, offset=i * 3 + 1) / 10)
            else:
                temps.append(None)

        return temps

    def _read_data(self, mode, data_type):
        self._send_command(_CMD_OPEN_ENDPOINT, mode)
        raw_data = self._send_command(_CMD_READ)
        self._send_command(_CMD_CLOSE_ENDPOINT)
        if tuple(raw_data[3:5]) != data_type:
            raise ExpectationNotMet('device returned incorrect data type')

        return raw_data[5:]

    def _send_command(self, command, data=()):
        # self.device.write expects buf[0] to be the report number or 0 if not used
        buf = bytearray(_REPORT_LENGTH + 1)

        # buf[1] when going out is always 08
        buf[1] = 0x08

        # Indexes for the buffer
        cmd_start = 2
        data_start = cmd_start + len(command)
        data_end = data_start + len(data)

        # Fill in the buffer
        buf[cmd_start:data_start] = command
        buf[data_start:data_end] = data

        self.device.clear_enqueued_reports()
        self.device.write(buf)

        res = self.device.read(_RESPONSE_LENGTH)
        while res[0] != 0x00:
            res = self.device.read(_RESPONSE_LENGTH)        
        buf = bytes(res)
        assert buf[1] == command[0], 'response does not match command'
        return buf

    @contextmanager
    def _wake_device_context(self):
        try:
            self._send_command(_CMD_WAKE)
            yield
        finally:
            self._send_command(_CMD_SLEEP)

    def _write_data(self, mode, data_type, data):
        self._read_data(mode, data_type)  # Will ensure we are writing the correct data type to avoid breakage

        self._send_command(_CMD_OPEN_ENDPOINT, mode)

        buf = bytearray(len(data) + len(data_type) + 4)
        buf[0: 2] = int.to_bytes(len(data) + 2, length=2, byteorder="little", signed=False)
        buf[4: 4 + len(data_type)] = data_type
        buf[4 + len(data_type):] = data

        self._send_command(_CMD_WRITE, buf)
        self._send_command(_CMD_CLOSE_ENDPOINT)

    def _fan_to_channel(self, fan):
        if self._has_pump:
            return fan
        else:
            # On devices without a pump, channel 0 is fan 1
            return fan - 1

    def _parse_channels(self, channel):
        if self._has_pump and channel == 'pump':
            return [0]
        elif channel == "fans":
            return [self._fan_to_channel(x) for x in range(1, _FAN_COUNT + 1)]
        elif channel.startswith("fan") and channel[3:].isnumeric() and 0 < int(channel[3:]) <= _FAN_COUNT:
            return [self._fan_to_channel(int(channel[3:]))]
        else:
            fan_names = ['fan' + str(i) for i in range(1, _FAN_COUNT + 1)]
            fan_names_part = '", "'.join(fan_names)
            if self._has_pump:
                fan_names.insert(0, "pump")
            raise ValueError(f'unknown channel, should be one of: "{fan_names_part}" or "fans"')

    def set_screen(self, channel, mode, value, **kwargs):
        """Not supported by this device."""
        raise NotSupportedByDevice



import pytest
from _testutils import MockHidapiDevice, Report

from collections import deque

from liquidctl.driver.rgb_fusion2 import RgbFusion2

# Sample data for 5702 controller from a Gigabyte Z490 Vision D
# https://github.com/liquidctl/liquidctl/issues/151#issuecomment-663213956
_INIT_5702_DATA = bytes.fromhex(
    'cc01000701000a00000000004954353730322d47494741425954452056312e30'
    '2e31302e30000000000102000200010002000100000102000001025700000000'
)
_INIT_5702_SAMPLE = Report(_INIT_5702_DATA[0], _INIT_5702_DATA[1:])

# Sample data for 8297 controller from a Gigabyte X570 Aorus Elite rev 1.0
# https://github.com/liquidctl/liquidctl/issues/151#issuecomment-663247422
# (note: original data had a trailing 0x61 byte, but that seems to be an artifact)
_INIT_8297_DATA = bytes.fromhex(
    '00010001010006000000000049543832393742582d4742583537300000000000'
    '0000000000000000000000000200010002000100000102000001978200000000'
)
_INIT_8297_SAMPLE = Report(_INIT_8297_DATA[0], _INIT_8297_DATA[1:])


@pytest.fixture
def mockRgbFusion2_5702Device():
    device = MockHidapiDevice(vendor_id=0x048d, product_id=0x5702, address='addr')
    dev = RgbFusion2(device, 'mock 5702 Controller')

    dev.connect()
    return dev


class Mock8297HidInterface(MockHidapiDevice):
    def get_feature_report(self, report_id, length):
        """Get a feature report emulating out of spec behavior of the device."""
        return super().get_feature_report(0, length)


@pytest.fixture
def mockRgbFusion2_8297Device():
    device = Mock8297HidInterface(vendor_id=0x048d, product_id=0x8297, address='addr')
    dev = RgbFusion2(device, 'mock 8297 Controller')

    dev.connect()
    return dev


def test_fusion2_5702_device_command_format(mockRgbFusion2_5702Device):
    mockRgbFusion2_5702Device.device.preload_read(_INIT_5702_SAMPLE)
    mockRgbFusion2_5702Device.initialize()
    mockRgbFusion2_5702Device.set_color(channel='sync', mode='off', colors=[])
    assert len(mockRgbFusion2_5702Device.device.sent) == 1 + 8 + 1

    for i, (report, data) in enumerate(mockRgbFusion2_5702Device.device.sent):
        assert report == 0xcc
        assert len(data) == 63  # TODO double check, more likely to be 64


def test_fusion2_5702_device_get_status(mockRgbFusion2_5702Device):
    assert mockRgbFusion2_5702Device.get_status() == []


def test_fusion2_5702_device_initialize_status(mockRgbFusion2_5702Device):
    mockRgbFusion2_5702Device.device.preload_read(_INIT_5702_SAMPLE)
    name, fw_version = mockRgbFusion2_5702Device.initialize()
    assert name[1] == "IT5702-GIGABYTE V1.0.10.0"
    assert fw_version[1] == '1.0.10.0'


def test_fusion2_5702_device_off_with_some_channel(mockRgbFusion2_5702Device):
    colors = [[0xff, 0, 0x80]]  # should be ignored
    mockRgbFusion2_5702Device.set_color(channel='led8', mode='off', colors=iter(colors))
    set_color, execute = mockRgbFusion2_5702Device.device.sent
    assert set_color.data[0:2] == [0x27, 0x80]
    assert set_color.data[10] == 0x01
    assert max(set_color.data[13:16]) == 0
    assert max(set_color.data[21:27]) == 0


def test_fusion2_5702_device_fixed_with_some_channel(mockRgbFusion2_5702Device):
    colors = [[0xff, 0, 0x80], [0x30, 0x30, 0x30]]  # second color should be ignored
    mockRgbFusion2_5702Device.set_color(channel='led7', mode='fixed', colors=iter(colors))
    set_color, execute = mockRgbFusion2_5702Device.device.sent
    assert set_color.data[0:2] == [0x26, 0x40]
    assert set_color.data[10] == 0x01
    assert set_color.data[13:16] == [0x80, 0x00, 0xff]
    assert max(set_color.data[21:27]) == 0


def test_fusion2_5702_device_pulse_with_some_channel_and_speed(mockRgbFusion2_5702Device):
    colors = [[0xff, 0, 0x80], [0x30, 0x30, 0x30]]  # second color should be ignored
    mockRgbFusion2_5702Device.set_color(channel='led3', mode='pulse', colors=iter(colors), speed='faster')
    set_color, execute = mockRgbFusion2_5702Device.device.sent

    assert set_color.data[0:2] == [0x22, 0x04]
    assert set_color.data[10] == 0x02
    assert set_color.data[13:16] == [0x80, 0x00, 0xff]
    assert set_color.data[21:27] == [0xe8, 0x03, 0xe8, 0x03, 0xf4, 0x01]


def test_fusion2_5702_device_flash_with_some_channel_and_speed(mockRgbFusion2_5702Device):
    colors = [[0xff, 0, 0x80], [0x30, 0x30, 0x30]]  # second color should be ignored
    mockRgbFusion2_5702Device.set_color(channel='led6', mode='flash', colors=iter(colors), speed='slowest')
    set_color, execute = mockRgbFusion2_5702Device.device.sent

    assert set_color.data[0:2] == [0x25, 0x20]
    assert set_color.data[10] == 0x03
    assert set_color.data[13:16] == [0x80, 0x00, 0xff]
    assert set_color.data[21:27] == [0x64, 0x00, 0x64, 0x00, 0x60, 0x09]


def test_fusion2_5702_device_double_flash_with_some_channel_and_speed_and_uppercase(mockRgbFusion2_5702Device):
    colors = [[0xff, 0, 0x80], [0x30, 0x30, 0x30]]  # second color should be ignored
    mockRgbFusion2_5702Device.set_color(channel='led5', mode='double-flash', colors=iter(colors),
                                        speed='ludicrous')
    set_color, execute = mockRgbFusion2_5702Device.device.sent

    assert set_color.data[0:2] == [0x24, 0x10]
    assert set_color.data[10] == 0x03
    assert set_color.data[13:16] == [0x80, 0x00, 0xff]
    assert set_color.data[21:27] == [0x64, 0x00, 0x64, 0x00, 0x40, 0x06]


def test_fusion2_5702_device_color_cycle_with_some_channel_and_speed(mockRgbFusion2_5702Device):
    colors = [[0xff, 0, 0x80]]  # should be ignored
    mockRgbFusion2_5702Device.set_color(channel='led4', mode='color-cycle', colors=iter(colors),
                                        speed='fastest')
    set_color, execute = mockRgbFusion2_5702Device.device.sent

    assert set_color.data[0:2] == [0x23, 0x08]
    assert set_color.data[10] == 0x04
    assert max(set_color.data[13:16]) == 0
    assert set_color.data[21:27] == [0x26, 0x02, 0xc2, 0x01, 0x00, 0x00]
    # TODO brightness


def test_fusion2_5702_device_common_behavior_in_all_set_color_writes(mockRgbFusion2_5702Device):
    colors = [[0xff, 0, 0x80]]
    for mode in ['off', 'fixed', 'pulse', 'flash', 'double-flash', 'color-cycle']:
        mockRgbFusion2_5702Device.device.sent = deque()
        mockRgbFusion2_5702Device.set_color(channel='led1', mode=mode, colors=iter(colors))
        set_color, execute = mockRgbFusion2_5702Device.device.sent

        assert execute.data[0:2] == [0x28, 0xff]
        assert max(execute.data[2:]) == 0


def test_fusion2_5702_device_sync_channel(mockRgbFusion2_5702Device):
    colors = [[0xff, 0, 0x80]]
    mockRgbFusion2_5702Device.set_color(channel='sync', mode='fixed', colors=iter(colors))
    assert len(mockRgbFusion2_5702Device.device.sent) == 8 + 1  # 8 × set + execute


def test_fusion2_5702_device_reset_all_channels(mockRgbFusion2_5702Device):
    mockRgbFusion2_5702Device.reset_all_channels()
    for addr, report in enumerate(mockRgbFusion2_5702Device.device.sent[:-1], 0x20):
        assert report.data[0:2] == [addr, 0]
        assert max(report.data[2:]) == 0
    execute = mockRgbFusion2_5702Device.device.sent[-1]
    assert execute.data[0:2] == [0x28, 0xff]
    assert max(execute.data[2:]) == 0


def test_fusion2_5702_device_invalid_set_color_arguments(mockRgbFusion2_5702Device):

    with pytest.raises(KeyError):
        mockRgbFusion2_5702Device.set_color('invalid', 'off', [])

    with pytest.raises(KeyError):
        mockRgbFusion2_5702Device.set_color('led1', 'invalid', [])

    with pytest.raises(ValueError):
        mockRgbFusion2_5702Device.set_color('led1', 'fixed', [])

    with pytest.raises(KeyError):
        mockRgbFusion2_5702Device.set_color('led1', 'pulse', [[0xff, 0, 0x80]], speed='invalid')


def test_fusion2_8297_device_initialize_status(mockRgbFusion2_8297Device):
    mockRgbFusion2_8297Device.device.preload_read(_INIT_8297_SAMPLE)
    name, fw_version = mockRgbFusion2_8297Device.initialize()

    assert name[1] == "IT8297BX-GBX570"
    assert fw_version[1] == '1.0.6.0'

    # other tests skipped, see Controller5702TestCase
